# SPDX-FileCopyrightText: 2023-present The Firebird Projects <www.firebirdsql.org>
#
# SPDX-License-Identifier: MIT
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/base/config.py
# DESCRIPTION:    Saturnin configuration
# CREATED:        6.2.2023
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2023 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""Code for configuration of Saturnin installation.

"""

from __future__ import annotations
from typing import Optional, List, Final
import sys
import os
import sysconfig
from pathlib import Path
from configparser import ConfigParser, ExtendedInterpolation
from firebird.base.config import DirectoryScheme, get_directory_scheme, Config, StrOption

#: filename for Saturnin configuration file
SATURNIN_CFG: Final[str] = 'saturnin.conf'
#: filename for Firebird configuration file
FIREBIRD_CFG: Final[str] = 'firebird.conf'
#: filename for logging configuration file
LOGGING_CFG: Final[str] = 'logging.conf'
#: Saturnin configuration file header
CONFIG_HDR: Final[str] = """;
; A configuration file consists of sections, each led by a [section] header, followed by
; key/value entries separated by = or : string. Section names are case sensitive but keys
; are not. Leading and trailing whitespace is removed from keys and values. Values can be
; omitted, in which case the key/value delimiter may also be left out. Values can also
; span multiple lines, as long as they are indented deeper than the first line of the value.
;
; Configuration files may include comments, prefixed by # and ; characters. Comments may
; appear on their own on an otherwise empty line, possibly indented.
;
; Values can contain ${section:option} format strings which refer to other values.
; If the section: part is omitted, interpolation defaults to the current section (and possibly
; the default values from the special DEFAULT section). Interpolation can span multiple levels.

"""

#: True if current platform is Windows
WINDOWS: Final[bool] = sys.platform == "win32"
#: True if current platform is based on MINGW
MINGW: Final[bool] = sysconfig.get_platform().startswith("mingw")

class SaturninScheme(DirectoryScheme):
    """Saturnin platform directory scheme.

    When SATURNIN_HOME environment variable is not set, and Saturnin resides in virtual
    environment that contains `home` subdirectory, it's set as Saturnin HOME directory.

    This `home` subdirectory is created by `saturnin create home` command on request.
    """
    def __init__(self):
        super().__init__('saturnin')
        if not self.has_home_env() and is_virtual():
            home_dir: Path = venv() / 'home'
            if home_dir.is_dir():
                os.environ['SATURNIN_HOME'] = str(home_dir)
        self.dir_map.update(get_directory_scheme('saturnin').dir_map)
        self.__pip_path: Path = 'pip'
        self.__pip_cmd: List[str] = ['pip']
        if is_virtual():
            root = venv()
            if WINDOWS:
                bin_path = root / "Scripts" if not MINGW else root / "bin"
                python_path = bin_path / "python.exe"
            else:
                bin_path = root / "bin"
                python_path = bin_path / "python"
            pip_path = bin_path / 'pip'
            if pip_path.is_file():
                self.__pip_path = pip_path
                self.__pip_cmd = [str(pip_path)]
            else:
                # No pip shortcut in venv, we must relly on python -m to run it, typical for pipx
                self.__pip_path = None
                self.__pip_cmd = [str(python_path), '-m', 'pip']
    def get_pip_cmd(self, *args) -> List[str]:
        """Returns list with command to run pip.

        Arguments:
           args: Arguments for pip
        """
        result = self.__pip_cmd.copy()
        result.extend(args)
        return result
    @property
    def recipes(self) -> Path:
        """Path to directory with recipe files.
        """
        return self.data / 'recipes'
    @property
    def pids(self) -> Path:
        """Path to directory with PID files for running daemons.
        """
        return self.run_data / 'pids'
    @property
    def site_services_toml(self) -> Path:
        """Saturnin service registry file.
        """
        return self.data / 'services.toml'
    @property
    def site_apps_toml(self) -> Path:
        """Saturnin application registry file.
        """
        return self.data / 'apps.toml'
    @property
    def site_oids_toml(self) -> Path:
        """Saturnin OID registry file.
        """
        return self.data / 'oids.toml'
    @property
    def site_conf(self) -> Path:
        """Saturnin site configuration file.
        """
        return self.config / SATURNIN_CFG
    @property
    def user_conf(self) -> Path:
        """Saturnin user configuration file.
        """
        return self.user_config / SATURNIN_CFG
    @property
    def firebird_conf(self) -> Path:
        """Firebird driver configuration file.
        """
        return self.config / FIREBIRD_CFG
    @property
    def logging_conf(self) -> Path:
        """Python logging configuration file.
        """
        return self.config / LOGGING_CFG
    @property
    def log_file(self) -> Path:
        """Saturnin log file.
        """
        return self.logs / 'saturnin.log'
    @property
    def history_file(self) -> Path:
        """Saturnin console command history file.
        """
        return self.data / 'saturnin.hist'
    @property
    def theme_file(self) -> Path:
        """Saturnin console theme file.
        """
        return self.config / 'theme.conf'
    @property
    def pip_path(self) -> Path:
        """Path to `pip`.
        """
        return self.__pip_path

class SaturninConfig(Config):
    """Saturnin (a Firebird Butler platform) configuration.
    """
    def __init__(self):
        super().__init__('saturnin')
        #: External editor
        self.editor: StrOption = StrOption('editor', "External editor",
                                           default=os.getenv('EDITOR'))

def is_virtual() -> bool:
    """Returns True if Saturnin runs in a virtual environtment.
    """
    # Check supports venv && virtualenv >= 20.0.0
    return getattr(sys, 'base_prefix', sys.prefix) != sys.prefix

def venv() -> Optional[Path]:
    """Path to Saturnin virtual environment.
    """
    return Path(sys.prefix) if is_virtual() else None

# Set SATURNIN_HOME if defined in virtual environment
if is_virtual():
    path: Path = venv() / 'home'
    if path.is_dir():
        os.environ['SATURNIN_HOME'] = str(path)

#: Active Saturnin directory scheme
directory_scheme: SaturninScheme = SaturninScheme()

#: Saturnin configuration object
saturnin_config: SaturninConfig = SaturninConfig()

parser: ConfigParser = ConfigParser(interpolation=ExtendedInterpolation())
parser.read([directory_scheme.site_conf, directory_scheme.user_conf])
if parser.has_section('saturnin'):
    saturnin_config.load_config(parser)
del parser



# SPDX-FileCopyrightText: 2020-present The Firebird Projects <www.firebirdsql.org>
#
# SPDX-License-Identifier: MIT
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/base/component.py
# DESCRIPTION:    Abstract base class for Saturnin Components
# CREATED:        3.12.2020
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2020 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""Saturnin abstract base class for Components.

Saturnin components are microservices, services or other software components that could be
executed on Saturnin platform, typically under supervision of Controller(s).
"""

from __future__ import annotations
from typing import Optional
from uuid import UUID
from abc import ABC, abstractmethod
from firebird.base.types import ZMQAddress
from firebird.base.config import Config, StrOption, UUIDOption

def create_config(_cls: Type[ComponentConfig], agent: UUID, name: str) -> ComponentConfig: # pragma: no cover
    """Returns newly created `ComponentConfig` instance.

    Intended to be used with `functools.partial` in `.ServiceDescriptor.config` definitions.

    Arguments:
      _cls:  Class for component configuration.
      agent: Component identification.
      name:  Name to be used by configuration class.
    """
    result: ComponentConfig = _cls(name)
    result.agent.value = agent
    return result

class ComponentConfig(Config):
    """Base Component configuration.

    Arguments:
       name: Component name (used as conf. file section name for component).
    """
    def __init__(self, name: str):
        super().__init__(name)
        #: Agent identification
        self.agent: UUIDOption = \
            UUIDOption('agent', "Agent identification. Do NOT change!")
        #: Logging ID for this component instance
        self.logging_id: StrOption = \
            StrOption('logging_id', "Logging ID for this component instance")

class Component(ABC):
    """Abstract base class for Saturnin Components.
    """
    @abstractmethod
    def initialize(self, config: ComponentConfig) -> None:
        """Verify configuration and assemble component structural parts.

        Arguments:
           config: Component configuration.
        """
    @abstractmethod
    def warm_up(self, ctrl_addr: Optional[ZMQAddress]) -> None:
        """Must initialize the `.ChannelManager` and connect component to control channel
        at provided address.

        Arguments:
            ctrl_addr: Controller's address for ICCP communication with component.
        """
    @abstractmethod
    def run(self) -> None:
        """Component execution (main loop).
        """



# SPDX-FileCopyrightText: 2021-present The Firebird Projects <www.firebirdsql.org>
#
# SPDX-License-Identifier: MIT
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/component/recipe.py
# DESCRIPTION:    Saturnin recipes
# CREATED:        11.3.2021
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2021 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________
# pylint: disable=R0913

"""Saturnin recipes


"""

from __future__ import annotations
from typing import Hashable
from uuid import UUID
from enum import Enum
from pathlib import Path
from dataclasses import dataclass
from configparser import ConfigParser, ExtendedInterpolation
from firebird.base.config import Config, EnumOption, StrOption, UUIDOption, PathOption
from firebird.base.types import Distinct, Error
from firebird.base.collections import Registry
from saturnin.base.config import directory_scheme

class RecipeType(Enum):
    """Recipe type.
    """
    SERVICE = 'service'
    BUNDLE = 'bundle'

class RecipeExecutionMode(Enum):
    """Recipe execution mode.
    """
    NORMAL = 'normal'
    DAEMON = 'daemon'

class SaturninRecipe(Config):
    """Saturnin recipe descriptor - configuration section.
    """
    def __init__(self):
        super().__init__('saturnin.recipe')
        #: Recipe type - `.RecipeType` enum.
        self.recipe_type: EnumOption = \
            EnumOption('recipe_type', RecipeType, "Type of recipe", required=True)
        #: Recipe execution mode - `.RecipeExecutionMode` enum.
        self.execution_mode: EnumOption = \
            EnumOption('execution_mode', RecipeExecutionMode, "Execution mode",
                       default=RecipeExecutionMode.NORMAL)
        #: Recipe executor (container). If not provided, the default executor according to recipe type is used.
        self.executor: PathOption = PathOption('executor', "Recipe executor.")
        #: Application (if any)
        self.application: UUIDOption = \
            UUIDOption('application', "Application UID")
        #: Description
        self.description: StrOption = \
            StrOption('description', "Recipe description", default="Not provided")

@dataclass(eq=True, order=False, frozen=True)
class RecipeInfo(Distinct): # pylint: disable=R0903
    """Dataclass recipe information record stored in recipe registry.

    Arguments:
        name: Recipe name
        recipe_type: Recipe type
        execution_mode: Recipe execution mode
        executor: Recipe executor
        application: Application to be used
        description: Recipe description
        filename: Path to recipe file
    """
    #: Recipe name
    name: str
    #: Recipe type
    recipe_type: RecipeType
    #: Recipe execution mode
    execution_mode: RecipeExecutionMode
    #: Recipe executor
    executor: Path
    #: Application to be used
    application: UUID
    #: Recipe description
    description: str
    #: Path to recipe file
    filename: Path
    def get_key(self) -> Hashable:
        "Returns service UID"
        return self.name

class RecipeRegistry(Registry):
    """Saturnin recipe registry.

    Holds `RecipeInfo` instances.
    """
    def load_from(self, directory: Path, *, ignore_errors: bool=False) -> None:
        """Populate registry from descriptors of installed recipes.

        Arguments:
          directory: Directory with recipe files.
          ignore_errors: When True, errors are ignored, otherwise `.Error` is raised.
        """
        recipe_cfg: SaturninRecipe = SaturninRecipe()
        cfg_file: ConfigParser = ConfigParser(interpolation=ExtendedInterpolation())
        for filename in directory.glob('*.cfg'):
            try:
                cfg_file.clear()
                cfg_file.read(filename)
            except Exception as exc:
                if ignore_errors:
                    continue
                raise Error(f"Failed to parse recipe '{filename}'") from exc
            try:
                recipe_cfg.clear()
                recipe_cfg.load_config(cfg_file)
            except Exception as exc:
                if ignore_errors:
                    continue
                raise Error(f"Malformed recipe '{filename}'") from exc
            self.store(RecipeInfo(filename.with_suffix('').name,
                                  recipe_cfg.recipe_type.value,
                                  recipe_cfg.execution_mode.value,
                                  recipe_cfg.executor.value,
                                  recipe_cfg.application.value,
                                  recipe_cfg.description.value,
                                  filename))

#: Saturnin recipe registry
recipe_registry: RecipeRegistry = RecipeRegistry()
if directory_scheme.recipes.is_dir():
    recipe_registry.clear()
    recipe_registry.load_from(directory_scheme.recipes)



# SPDX-FileCopyrightText: 2020-present The Firebird Projects <www.firebirdsql.org>
#
# SPDX-License-Identifier: MIT
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/component/registry.py
# DESCRIPTION:    Component registration and discovery
# CREATED:        4.12.2020
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2020 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""Saturnin component registration and discovery.

Services are registered as entry points for their `.ServiceDescriptor` - i.e. the instance
of `.ServiceDescriptor` for installed service is returned by `EntryPoint.load()`.

The default group for service registration is `saturnin.service`, but it's possible to
install additional service discovery iterators.

Custom service iterator must be a generator that accepts optional `uid` in string format,
and yileds `importlib.metadata.EntryPoint` instances. If `uid` is specified, it must return
only `EntryPoint` only for service with given `uid`, otherwise it should return all services.

Note:

  Custom iterator can return objects that are not `~importlib.metadata.EntryPoint` instances,
  but they MUST implement `load()` method that will return `.ServiceDescriptor` instance.

Custom iterators must be registered as entry points in `saturnin.service.iterator` group.
"""

from __future__ import annotations
from typing import List, Dict, Hashable, Optional, Any
from functools import partial
from itertools import chain
from contextlib import suppress
from uuid import UUID
from toml import dumps, loads
from firebird.base.types import Distinct, load
from firebird.base.collections import Registry
from saturnin.base import directory_scheme, ServiceDescriptor, Error
from saturnin.lib.metadata import iter_entry_points, get_entry_point_distribution

class ServiceInfo(Distinct): # pylint: disable=R0902
    """Information about service stored in  `.ServiceRegistry`.

    Arguments:
        uid: Service UID
        name: Service name
        version: Service version
        vendor: Service vendor UID
        classification: Service classification
        description: Service description
        facilities: List of service facilities
        api: List of interfaces provided by service
        factory: Service factory specification (entry point)
        descriptor: Service descriptor specification (entry point)
        distribution: Installed distribution package that contains this service
    """
    def __init__(self, *, uid: UUID, name: str, version: str, vendor: UUID,
                 classification: str, description: str, facilities: List[str],
                 api: List[UUID], factory: str, descriptor: str, distribution: str):
        self.__desc_obj: Any = None
        self.__fact_obj: Any = None
        #: Service UID
        self.uid: UUID = uid
        #: Service name
        self.name: str = name
        #: Service version
        self.version: str = version
        #: Service vendor UID
        self.vendor: UUID = vendor
        #: Service classification
        self.classification: str = classification
        #: Service description
        self.description: str = description
        #: List of service facilities
        self.facilities: List[str] = facilities
        #: List of interfaces provided by service
        self.api: List[UUID] = api
        #: Service factory specification (entry point)
        self.factory: str = factory
        #: Service descriptor specification (entry point)
        self.descriptor: str = descriptor
        #: Installed distribution package that contains this service
        self.distribution: str = distribution
    def get_key(self) -> Hashable:
        "Returns service UID"
        return self.uid
    def as_toml_dict(self) -> Dict:
        """Returns dictionary with instance data suitable for storage in TOML format
        (values that are not of basic type are converted to string).
        """
        return {'uid': str(self.uid),
                'name': self.name,
                'version': self.version,
                'vendor': str(self.vendor),
                'classification': self.classification,
                'description': self.description,
                'facilities': self.facilities,
                'api': [str(x) for x in self.api],
                'factory': self.factory,
                'descriptor': self.descriptor,
                'distribution': self.distribution,
                }
    def __get_descriptor_obj(self) -> ServiceDescriptor:
        """Service descriptor object. If it's not assigned directly, then it's loaded
        using `.descriptor` on first access.
        """
        if self.__desc_obj is None:
            self.__desc_obj = load(self.descriptor)
        return self.__desc_obj
    def __set_descriptor_obj(self, value: Optional[ServiceDescriptor]) -> None:
        "Property setter"
        self.__desc_obj = value
    descriptor_obj = property(__get_descriptor_obj, __set_descriptor_obj, None,
                              __get_descriptor_obj.__doc__)
    def __get_factory_obj(self) -> Any:
        """Service factory object. If it's not assigned directly, then it's loaded
        using `.factory` on first access.
        """
        if self.__fact_obj is None:
            self.__fact_obj = load(self.factory)
        return self.__fact_obj
    def __set_factory_obj(self, value: Optional[Any]) -> None:
        "Property setter"
        self.__fact_obj = value
    factory_obj = property(__get_factory_obj, __set_factory_obj, None, __get_factory_obj.__doc__)

class ServiceRegistry(Registry):
    """Saturnin service registry.

    Holds `.ServiceInfo` instances.

    It is used in two modes:

    1. In full saturnin deployment, the information about services is loaded from TOML file.
       Service descriptors and factories are loaded on demand.
    2. In standalone service/bundle mode, service information including service desciptors
       and factories is stored directly by executor script, so there is no dynamic discovery
       and the whole could be compiled with Nutika.
    """
    def add(self, descriptor: ServiceDescriptor, factory: Any, distribution: str) -> None:
        """Direct service registration. Used by systems that does not allow dynamic discovery,
        for example programs compiled by Nuitka.

        Arguments:
            descriptor: Service descriptor
            factory: Service factory
            distribution: Distribution package name with service
        """
        kwargs = {}
        kwargs['distribution'] = distribution
        kwargs['descriptor'] = ''
        kwargs['uid'] = descriptor.agent.uid
        kwargs['name'] = descriptor.agent.name
        kwargs['version'] = descriptor.agent.version
        kwargs['vendor'] = descriptor.agent.vendor_uid
        kwargs['classification'] = descriptor.agent.classification
        kwargs['description'] = descriptor.description
        kwargs['facilities'] = descriptor.facilities
        kwargs['api'] = [x.get_uid() for x in descriptor.api]
        kwargs['factory'] = descriptor.factory
        svc_info = ServiceInfo(**kwargs) # pylint: disable=E1125
        svc_info.descriptor_obj = descriptor
        svc_info.factory_obj = factory
        self.store(svc_info)
    def load_from_installed(self, *, ignore_errors: bool=False) -> None:
        """Populate registry from descriptors of installed services.

        Arguments:
          ignore_errors: When True, errors are ignored, otherwise `.Error` is raised.
        """
        for entry in chain.from_iterable([i() for i in _iterators]):
            kwargs = {}
            dist = get_entry_point_distribution(entry)
            kwargs['distribution'] = dist if dist is None else dist.metadata['name']
            kwargs['descriptor'] = entry.value
            try:
                desc: ServiceDescriptor = entry.load()
            except Exception as exc:
                if ignore_errors:
                    continue
                raise Error(f"Failed to load service '{entry.name}' "
                            f"from '{kwargs['distribution']}'") from exc
            kwargs['uid'] = desc.agent.uid
            kwargs['name'] = desc.agent.name
            kwargs['version'] = desc.agent.version
            kwargs['vendor'] = desc.agent.vendor_uid
            kwargs['classification'] = desc.agent.classification
            kwargs['description'] = desc.description
            kwargs['facilities'] = desc.facilities
            kwargs['api'] = [x.get_uid() for x in desc.api]
            kwargs['factory'] = desc.factory
            try:
                svc_info = ServiceInfo(**kwargs) # pylint: disable=E1125
            except Exception as exc:
                if ignore_errors:
                    continue
                raise Error(f"Malformed service descriptor for '{entry.name}' "
                            f"from '{kwargs['distribution']}'") from exc
            with suppress(ValueError):
                self.store(svc_info)
    def load_from_toml(self, toml: str, *, ignore_errors: bool=False) -> None:
        """Populate registry from TOML document.

        Arguments:
          toml: TOML document (as created by `as_toml` method).
          ignore_errors: When True, errors are ignored, otherwise `.Error` is raised.
        """
        data = loads(toml)
        self.clear()
        for uid, kwargs in data.items():
            try:
                kwargs['uid'] = UUID(kwargs['uid'])
                kwargs['vendor'] = UUID(kwargs['vendor'])
                kwargs['api'] = [UUID(x) for x in kwargs['api']]
                svc_info = ServiceInfo(**kwargs)
            except Exception as exc:
                if ignore_errors:
                    continue
                raise Error(f"Malformed service data for '{uid}' "
                            f"from '{kwargs['distribution']}'") from exc
            self.store(svc_info)
    def as_toml(self) -> str:
        """Returns registry content as TOML document.
        """
        nodes = {str(node.uid): node.as_toml_dict() for node in self._reg.values()}
        return dumps(nodes)
    def load(self) -> None:
        "Read information about installed services from previously saved TOML file."
        if directory_scheme.site_services_toml.is_file():
            service_registry.load_from_toml(directory_scheme.site_services_toml.read_text())
    def save(self) -> None:
        "Save information about installed services to TOML file."
        directory_scheme.site_services_toml.write_text(service_registry.as_toml())
    def get_by_name(self, name: str, default: Any=None) -> Distinct:
        """Get service by its name.

        Arguments:
            name: Service name.
            default: Default value returned when service is not found.
        """
        return self.find(lambda x: x.name == name, default=default)

# Default service registration
_iterators = [partial(iter_entry_points, 'saturnin.service')]

# Load custom service iterators
for i in iter_entry_points('saturnin.service.iterator'):
    _iterators.append(i.load())

#: Saturnin service registry
service_registry: ServiceRegistry = ServiceRegistry()
service_registry.load()



#coding:utf-8
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/lib/data/onepipe.py
# DESCRIPTION:    Base class for Saturnin data provider and consumer microservices
# CREATED:        14.12.2020
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2020 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________
# pylint: disable=R0903, R0902

"""Saturnin base classes for data provider and consumer microservices

This is extended description.
"""

from __future__ import annotations
from typing import cast, Union, Final
from functools import partial
import uuid
import zmq
from firebird.base.config import (StrOption, EnumOption, IntOption, BoolOption,
    ZMQAddressOption, MIMEOption)
from saturnin.base import (Error, StopError, Direction, SocketMode, PipeSocket, Outcome,
     ZMQAddress, MIME, ComponentConfig, Channel, Session, Message, DealerChannel,
     ServiceDescriptor)
from saturnin.component.micro import MicroService
from saturnin.protocol.fbdp import (ErrorCode, FBDPServer, FBDPClient, FBDPSession,
    FBDPMessage)

#: Channel & endpoint name
PIPE_CHN: Final[str] = 'pipe'

class BaseDataPipeConfig(ComponentConfig):
    """Base data provider/consumer microservice configuration.
    """
    def __init__(self, name: str):
        super().__init__(name)
        #: Stop service when pipe is closed
        self.stop_on_close: BoolOption = \
            BoolOption('stop_on_close', "Stop service when pipe is closed", default=True)
        #: Data Pipe Identification
        self.pipe: StrOption = \
            StrOption('pipe', "Data Pipe Identification", required=True)
        #: Data Pipe endpoint address
        self.pipe_address: ZMQAddressOption = \
            ZMQAddressOption('pipe_address', "Data Pipe endpoint address", required=True)
        #: Data Pipe Mode
        self.pipe_mode: EnumOption = \
            EnumOption('pipe_mode', SocketMode, "Data Pipe Mode", required=True)
        #: Pipe data format specification
        self.pipe_format: MIMEOption = \
            MIMEOption('pipe_format', "Pipe data format specification")
        #: Data batch size
        self.batch_size: IntOption = \
            IntOption('batch_size', "Data batch size", required=True, default=50)
        #: READY message schedule interval in milliseconds
        self.ready_schedule_interval: IntOption = \
            IntOption('ready_schedule_interval',
                      "READY message schedule interval in milliseconds", required=True,
                      default=1000)
    def validate(self) -> None:
        """Extended validation.

        - `pipe_format` is required for CONNECT `pipe_mode`.
        """
        super().validate()
        if self.pipe_mode.value is SocketMode.CONNECT and self.pipe_format.value is None:
            raise Error("'pipe_format' required for CONNECT pipe mode.")

class DataProviderConfig(BaseDataPipeConfig):
    """Base data provider microservice configuration.
    """

class DataConsumerConfig(BaseDataPipeConfig):
    """Base data consumer microservice configuration.
    """

class BaseDataPipeMicro(MicroService):
    """Base data provider/consumer microservice.

    Descendant classes should override:

    - `handle_accept_client` to validate client request and aquire resources associated
      with pipe.
    - `handle_produce_data` to produce data for outgoing DATA message. PRODUCER only.
    - `handle_accept_data` to process received data. CONSUMER only.
    - `handle_pipe_closed` to release resource assiciated with pipe.
    """
    def __init__(self, zmq_context: zmq.Context, descriptor: ServiceDescriptor, *,
                 peer_uid: uuid.UUID=None):
        """
        Arguments:
            zmq_context: ZeroMQ Context.
            descriptor: Service descriptor.
            peer_uid: Peer ID, `None` means that newly generated UUID type 1 should be used.
        """
        super().__init__(zmq_context, descriptor, peer_uid=peer_uid)
        #: Pipe socket this service handles if operated as server (bind). Must be set
        #: in descendant class.
        #: For PROVIDER it's `.PipeSocket.OUTPUT`, for CONSUMER it's `.PipeSocket.INPUT`
        self.server_socket: PipeSocket = None
        #: FDBP protocol handler (server or client)
        self.protocol: Union[FBDPServer, FBDPClient] = None
        # Next members are set in initialize()
        #: [Configuration] Whether service should stop when pipe is closed
        self.stop_on_close: bool = None
        #: [Configuration] Data Pipe Identification
        self.pipe: str = None
        #: [Configuration] Data Pipe Mode
        self.pipe_mode: SocketMode = None
        #: [Configuration] Data Pipe endpoint address
        self.pipe_address: ZMQAddress = None
        #: [Configuration] Pipe data format specification
        self.pipe_format: MIME = None
        #: [Configuration] Data batch size
        self.batch_size: int = None
        #: [Configuration] READY message schedule interval in milliseconds
        self.ready_schedule_interval: int = None
    def initialize(self, config: BaseDataPipeConfig) -> None:
        """Verify configuration and assemble component structural parts.
        """
        super().initialize(config)
        # Configuration
        self.stop_on_close = config.stop_on_close.value
        self.pipe: str = config.pipe.value
        self.pipe_mode: SocketMode = config.pipe_mode.value
        self.pipe_address: ZMQAddress = config.pipe_address.value
        self.pipe_format: MIME = config.pipe_format.value
        self.batch_size: int = config.batch_size.value
        self.ready_schedule_interval: int = config.ready_schedule_interval.value
        # Set up FBDP protocol
        if self.pipe_mode == SocketMode.BIND:
            # server
            self.protocol = FBDPServer()
            self.protocol.on_exception = self.handle_exception
            self.protocol.on_accept_client = self.handle_accept_client
            self.protocol.on_schedule_ready = self.handle_schedule_ready
            # We have an endpoint to bind
            self.endpoints[PIPE_CHN] = [self.pipe_address]
        else:
            # client
            self.protocol = FBDPClient()
        # common parts
        self.protocol.log_context = self.logging_id
        self.protocol.batch_size = self.batch_size
        self.protocol.on_pipe_closed = self.handle_pipe_closed
        self.protocol.on_produce_data = self.handle_produce_data
        self.protocol.on_accept_data = self.handle_accept_data
        # Create pipe channel
        self.mngr.create_channel(DealerChannel, PIPE_CHN, self.protocol, wait_for=Direction.IN)
    def aquire_resources(self) -> None:
        """Aquire resources required by component:

        1. If `.pipe_mode` is `~SocketMode.CONNECT`, it will connect to the end of the pipe
           that is the inverse of `.server_socket`.
        """
        # Connect to the data pipe
        if self.pipe_mode == SocketMode.CONNECT:
            chn: Channel = self.mngr.channels[PIPE_CHN]
            session = chn.connect(self.pipe_address)
            # OPEN the data pipe connection, this also fills session attributes
            # PRODUCER client must attach to INPUT, CONSUMER client must attach to OUTPUT
            client_socket = PipeSocket.INPUT if self.server_socket is PipeSocket.OUTPUT \
                else PipeSocket.OUTPUT
            cast(FBDPClient, chn.protocol).send_open(chn, session, self.pipe,
                                                     client_socket, self.pipe_format)
    def release_resources(self) -> None:
        """Release resources aquired by component:

        1. CLOSE all active pipe sessions.
        """
        # CLOSE all active data pipe sessions
        chn: Channel = self.mngr.channels[PIPE_CHN]
        # send_close() will discard session, so we can't iterate over sessions.values() directly
        for session in list(chn.sessions.values()):
            # We have to report error here, because normal is to close pipes before
            # shutdown is commenced. Mind that service shutdown could be also caused by error!
            cast(FBDPServer, chn.protocol).send_close(chn, session, ErrorCode.ERROR)
    def handle_exception(self, channel: Channel, session: Session, msg: Message, # pylint: disable=W0613
                         exc: Exception) -> None:
        """Event handler called by `.handle_msg()` on exception in message handler.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with connection.
            msg:     Message.
            exc:     Exception.

        Sets service outcome to ERROR and notes exception as details.
        """
        self.outcome = Outcome.ERROR
        self.details = exc
    # FBDP server only
    def handle_accept_client(self, channel: Channel, session: FBDPSession) -> None: # pylint: disable=W0613
        """Event handler executed when client connects to the data pipe via OPEN message.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with client.

        The session attributes `~.FBDPSession.pipe`, `~.FBDPSession.socket`,
        `~.FBDPSession.data_format` and `~.FBDPSession.params` contain information sent by
        client, and the event handler validates the request.

        If request should be rejected, it raises the `.StopError` exception with `~Error.code`
        attribute containing the `~saturnin.protocol.fbdp.ErrorCode` to be returned in
        CLOSE message.

        Important:
            Base implementation validates pipe identification and pipe socket, and converts
            data format from string to MIME (in session).

            The descendant class that overrides this method must call `super` as first
            action.
        """

        if session.pipe != self.pipe:
            raise StopError(f"Unknown data pipe '{session.data_pipe}'",
                            code = ErrorCode.PIPE_ENDPOINT_UNAVAILABLE)
        # We're server, so clients can only attach to our server_socket
        if session.socket is not self.server_socket:
            raise StopError(f"'{session.socket}' socket not available",
                            code = ErrorCode.PIPE_ENDPOINT_UNAVAILABLE)
        # We work with MIME formats, so we'll convert the format specification to MIME
        session.data_format = MIME(session.data_format)
    def handle_schedule_ready(self, channel: Channel, session: FBDPSession) -> None:
        """Event handler executed in order to send the READY message to the client later.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with client.

        The event handler may cancel the transmission by raising the `.StopError` exception
        with `~Error.code` attribute containing the `~saturnin.protocol.fbdp.ErrorCode` to be
        returned in CLOSE message.

        Important:
            The base implementation schedules `~.FBDPServer.resend_ready()` according to
            `.ready_schedule_interval` configuration option.
        """
        self.schedule(partial(cast(FBDPServer, channel.protocol).resend_ready,
                              channel, session),
                      self.ready_schedule_interval)
    # FBDP common
    def handle_produce_data(self, channel: Channel, session: FBDPSession, msg: FBDPMessage) -> None:
        """Event handler executed to store data into outgoing DATA message.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with client.
            msg: DATA message that will be sent to client.

        The event handler must store the data in `msg.data_frame` attribute. It may also
        set ACK-REQUEST flag and `type_data` attribute.

        The event handler may cancel the transmission by raising the `.StopError` exception
        with `code` attribute containing the `~saturnin.protocol.fbdp.ErrorCode` to be
        returned in CLOSE message.

        Note:
            To indicate end of data, raise StopError with ErrorCode.OK code.

        Note:
            Exceptions are handled by protocol, but only StopError is checked for protocol
            ErrorCode. As we want to report INVALID_DATA properly, we have to convert
            UnicodeError into StopError.

        Important:
            The base implementation simply raises `.StopError` with `ErrorCode.OK` code,
            so the descendant class must override this method without `super` call.
        """
        raise StopError('OK', code=ErrorCode.OK)
    def handle_accept_data(self, channel: Channel, session: FBDPSession, data: bytes) -> None:
        """Event handler executed to process data received in DATA message.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with client.
            data: Data received from client.

        The event handler may cancel the transmission by raising the `.StopError` exception
        with `~Error.code` attribute containing the `~saturnin.protocol.fbdp.ErrorCode` to
        be returned in CLOSE message.

        Note:
            The ACK-REQUEST in received DATA message is handled automatically by protocol.

        Important:
            The base implementation simply raises `.StopError` with `ErrorCode.OK` code,
            so the descendant class must override this method without `super` call.
        """
        raise StopError('OK', code=ErrorCode.OK)
    def handle_pipe_closed(self, channel: Channel, session: FBDPSession, msg: FBDPMessage, # pylint: disable=W0613
                           exc: Exception=None) -> None:
        """Event handler executed when CLOSE message is received or sent, to release any
        resources associated with current transmission.

        Arguments:
            channel: Channel associated with data pipe.
            session: Session associated with peer.
            msg: Received/sent CLOSE message.
            exc: Exception that caused the error.

        Important:
            The base implementation does only two actions:

            - If exception is provided, sets service execution outcome to ERROR
              and notes exception in details.
            - Stops the service if `.stop_on_close` is True.

            The descendant class that overrides this method must call `super`.
        """
        # FDBP converts exceptions raised in our event handler to CLOSE messages, so
        # here is the central place to handle errors in data pipe processing.
        # Note problem in service execution outcome
        if exc is not None:
            self.outcome = Outcome.ERROR
            self.details = exc
        #
        if self.stop_on_close:
            self.stop.set()


class DataProviderMicro(BaseDataPipeMicro):
    """Base data provider microservice.

    Descendant classes should override:

    - `~.BaseDataPipeMicro.handle_accept_client` to validate client request and aquire
      resources associated with pipe.
    - `~.BaseDataPipeMicro.handle_produce_data` to produce data for outgoing DATA message.
    - `~.BaseDataPipeMicro.handle_pipe_closed` to release resource assiciated with pipe.
    """
    def initialize(self, config: DataProviderConfig) -> None:
        """Verify configuration and assemble component structural parts.
        """
        super().initialize(config)
        self.server_socket = PipeSocket.OUTPUT
        # High water mark optimization
        chn: Channel = self.mngr.channels[PIPE_CHN]
        chn.sock_opts['rcvhwm'] = 5
        chn.sock_opts['sndhwm'] = int(self.batch_size / 2) + 5

class DataConsumerMicro(BaseDataPipeMicro):
    """Base data provider microservice.

    Descendant classes should override:

    - `~.BaseDataPipeMicro.handle_accept_client` to validate client request and aquire
      resources associated with pipe.
    - `~.BaseDataPipeMicro.handle_accept_data` to process received data.
    - `~.BaseDataPipeMicro.handle_pipe_closed` to release resource assiciated with pipe.
    """
    def initialize(self, config: DataConsumerConfig) -> None:
        """Verify configuration and assemble component structural parts.
        """
        super().initialize(config)
        self.server_socket = PipeSocket.INPUT
        # High water mark optimization
        chn: Channel = self.mngr.channels[PIPE_CHN]
        chn.sock_opts['rcvhwm'] = int(self.batch_size / 2) + 5
        chn.sock_opts['sndhwm'] = 5



# SPDX-FileCopyrightText: 2023-present The Firebird Projects <www.firebirdsql.org>
#
# SPDX-License-Identifier: MIT
#
#   PROGRAM/MODULE: saturnin
#   FILE:           saturnin/_scripts/completers.py
#   DESCRIPTION:    CLI command completers
#   CREATED:        19.02.2023
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Initial code is based on https://github.com/click-contrib/click-repl
#
# Contributor(s): Pavel Císař (initial code)
#                 ______________________________________
# pylint: disable=W0613

"""CLI command completers.

For use with Typer/Click arguments and options (parameter `autocompletion`).
"""

from __future__ import annotations
from typing import List
from pathlib import Path
from firebird.uuid import oid_registry
from saturnin.component.recipe import recipe_registry
from saturnin.component.registry import service_registry
from saturnin.component.apps import application_registry

def get_first_line(text: str) -> str:
    """Returns first non-empty line from argument.
    """
    return text.strip().split('\n')[0]

def oid_completer(ctx, args, incomplete) -> List:
    """Click completer for OIDs. Returns both, UUID and OID names.
    """
    result = [(str(oid.uid)) for oid in oid_registry.values()]
    result.extend(oid.full_name for oid in oid_registry.values())
    return result

def service_completer(ctx, args, incomplete) -> List:
    """Click completer for Saturnin services.
    """
    result = [(str(svc.uid)) for svc in service_registry.values()]
    result.extend(svc.name for svc in service_registry.values())
    return result

def application_completer(ctx, args, incomplete) -> List:
    """Click completer for Saturnin applications.
    """
    result = [(str(app.uid)) for app in application_registry.values()]
    result.extend(app.name for app in application_registry.values())
    return result

def recipe_completer(ctx, args, incomplete) -> List:
    """Click completer for Saturnin recipes.
    """
    return [(recipe.name, get_first_line(recipe.description))
            for recipe in recipe_registry.values()]

def path_completer(ctx, args, incomplete) -> List:
    """Click completer for Path values.
    """
    base_path = Path(incomplete)
    while not base_path.exists():
        base_path = base_path.parent
    return [str(path) for path in base_path.glob('*')]



# SPDX-FileCopyrightText: 2019-present The Firebird Projects <www.firebirdsql.org>
#
# SPDX-License-Identifier: MIT
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/component/micro.py
# DESCRIPTION:    Base module for implementation of Firebird Butler Microservices
# CREATED:        22.4.2019
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2019 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________.
# pylint: disable=R0902, R1702

"""Saturnin base module for implementation of Firebird Butler Microservices.
"""

from __future__ import annotations
from typing import Union, Dict, List, Optional, Callable, cast, Final
import os
import platform
import threading
import uuid
from contextlib import suppress
from weakref import proxy
from time import monotonic_ns
from heapq import heappush, heappop
import zmq
from firebird.base.types import Conjunctive
from firebird.base.trace import TracedMixin
from saturnin.base import (ZMQAddress, Component, PeerDescriptor, ServiceDescriptor,
     ServiceError, Direction, State, Outcome, ChannelManager, Channel, PairChannel,
     ComponentConfig, ConfigProto, PrioritizedItem)
from saturnin.protocol.iccp import ICCPComponent

#: Service control channel name
SVC_CTRL: Final[str] = 'iccp'

class MicroService(Component, TracedMixin, metaclass=Conjunctive): # pylint: disable=E1139
    """Saturnin Component for Firebird Burler Microservices.
    """
    def __init__(self, zmq_context: zmq.Context, descriptor: ServiceDescriptor, *,
                 peer_uid: uuid.UUID=None):
        """
        Arguments:
            zmq_context: ZeroMQ Context.
            descriptor: Service descriptor.
            peer_uid: Peer ID, `None` means that newly generated UUID type 1 should be used.
        """
        self._heap: List = []
        #: Service execution outcome
        self.outcome: Outcome = Outcome.UNKNOWN
        #: Service execution outcome details
        self.details: Union[Exception, List[str]] = None
        #: Service internal state
        self.state: State = State.UNKNOWN_STATE
        #: Event to stop the component
        self.stop: threading.Event = threading.Event()
        #: ChannelManager instance.
        self.mngr: ChannelManager = ChannelManager(zmq_context)
        self.mngr.log_context = proxy(self)
        #: Dictionary with endpoints to which the component binds.
        #: Key is channel name, value is list of ZMQAddress instances.
        #: Initially empty.
        self.endpoints: Dict[str, List[ZMQAddress]] = {}
        #: Service desriptor.
        self.descriptor: ServiceDescriptor = descriptor
        #: Peer descriptor for this component.
        self.peer: PeerDescriptor = PeerDescriptor(uuid.uuid1() if peer_uid is None else peer_uid,
                                                   os.getpid(), platform.node())
    def __str__(self):
        return self.logging_id
    __repr__ = __str__
    def handle_stop_component(self, exc: Exception=None) -> None:
        """ICCP event handler. Called when commponent should stop its operation.
        It stops the component by setting the `~Component.stop` event.

        Arguments:
           exc: Exception that describes the reason why component should stop. If not
                provided, the component should stop on controller's request.
        """
        if exc is not None:
            self.outcome = Outcome.ERROR
            self.details = exc
        self.stop.set()
    def handle_config_request(self, config: ConfigProto) -> None:
        """ICCP event handler. Called when controller requested reconfiguration.

        Must raise an exception if configuration fails for any reason.

        By default, the component does not support run-time configuration, so it raises
        NotImplementedError.

        Arguments:
           config: New configuration provided by controller.
        """
        raise NotImplementedError("Service does not support run-time configuration")
    def schedule(self, action: Callable, after: int) -> None:
        """Schedule action to be executed after specified time.

        Action is executed in `run()` main I/O loop not sooner than after specified number
        of milliseconds from time when `schedule()` is called. However, the actual delay
        could be longer than specified (depends on time spent in message handlers and other
        factors).

        Arguments:
            action: Callable (without arguments) to be executed. Use `functools.partial`
                    if callable requires arguments.
            after:  Delay in milliseconds.
        """
        heappush(self._heap, PrioritizedItem(monotonic_ns() + (after * 1000000), action))
    def get_timeout(self) -> int:
        """Returns timeout to next scheduled action.
        """
        if not self._heap:
            return 1000
        back = []
        i = len(self._heap)
        now = monotonic_ns()
        while self._heap and (item := heappop(self._heap)).priority < now:
            back.append(item)
        if len(back) != i:
            heappush(self._heap, item)
        for value in back:
            heappush(self._heap, value)
        return max(int((item.priority - now) / 1000000), 0)
    def run_scheduled(self) -> None:
        """Run scheduled actions.
        """
        while self._heap:
            item = heappop(self._heap)
            if item.priority < monotonic_ns():
                item.item()
            else:
                heappush(self._heap, item)
                break
    def initialize(self, config: ComponentConfig) -> None:
        """Verify configuration and assemble component structural parts.

        Arguments:
            config: Service configuration.
        """
        config.validate() # Fail early!
        if config.logging_id.value is not None:
            self._logging_id_ = config.logging_id.value # pylint: disable=W0201
    def bind_endpoints(self) -> None:
        """Bind endpoints used by component.
        """
        for name, addr_list in self.endpoints.items():
            chn: Channel = self.mngr.channels.get(name)
            for i, addr in enumerate(addr_list):
                #self.endpoints[name][i] = chn.bind(addr)
                addr_list[i] = chn.bind(addr)
    def aquire_resources(self) -> None:
        """Aquire resources required by component (open files, connect to other services etc.).

        Must raise an exception when resource aquisition fails.
        """
    def release_resources(self) -> None:
        """Release resources aquired by component (close files, disconnect from other services etc.)
        """
    def start_activities(self) -> None:
        """Start normal component activities.

        Must raise an exception when start fails.
        """
    def stop_activities(self) -> None:
        """Stop component activities.
        """
    def warm_up(self, ctrl_addr: Optional[ZMQAddress]) -> None:
        """Initializes the `.ChannelManager` and connects component to control channel.
        """
        if ctrl_addr is not None:
            # Service control channel
            iccp = ICCPComponent(with_traceback=__debug__)
            iccp.on_stop_component = self.handle_stop_component
            iccp.on_config_request = self.handle_config_request
            chn: PairChannel = self.mngr.create_channel(PairChannel, SVC_CTRL, iccp,
                                                        wait_for=Direction.IN,
                                                        sock_opts={'rcvhwm': 5,
                                                                   'sndhwm': 5,})
            chn.protocol.log_context = self.logging_id
        self.mngr.warm_up()
        if ctrl_addr is not None:
            chn.connect(ctrl_addr)
            if not chn.can_send():
                raise ServiceError("Broken component control channel")
        try:
            self.bind_endpoints()
            self.aquire_resources()
            self.start_activities()
        except Exception as exc:
            if ctrl_addr is not None:
                chn.send(cast(ICCPComponent, chn.protocol).error_msg(exc), chn.session)
            self.mngr.shutdown()
            raise
        else:
            if ctrl_addr is not None:
                chn.send(cast(ICCPComponent, chn.protocol).ready_msg(self.peer, self.endpoints),
                         chn.session)
            self.state = State.READY
    def run(self) -> None:
        """Component execution (main loop).
        """
        self.state = State.RUNNING
        ctrl_chn: PairChannel = self.mngr.channels.get(SVC_CTRL)
        try:
            while not self.stop.is_set():
                events = self.mngr.wait(self.get_timeout())
                if events:
                    # Messages from service control channel have top priority
                    if ctrl_chn in events:
                        ctrl_chn.receive()
                        if self.stop.is_set():
                            continue # stop quickly
                    # Channels waiting for output have precedence
                    if self.mngr.has_pollout():
                        for chn, event in events.items():
                            if Direction.OUT in event:
                                chn.on_output_ready(chn)
                    # Now process incomming messages
                    for chn, event in events.items():
                        if Direction.IN in event:
                            chn.receive()
                # Now it's time for scheduled actions
                self.run_scheduled()
            # Gracefully stop the service
            self.state = State.STOPPED
            self.stop_activities()
            self.release_resources()
            if self.outcome is Outcome.UNKNOWN:
                self.outcome = Outcome.OK
            ctrl_chn.send(cast(ICCPComponent, ctrl_chn.protocol).finished_msg(self.outcome,
                                                                              self.details),
                          ctrl_chn.session)
            self.mngr.shutdown()
            self.state = State.FINISHED
        except Exception as exc: # pylint: disable=W0703
            self.state = State.ABORTED
            with suppress(Exception):
                # try send report to controller
                ctrl_chn.send(cast(ICCPComponent, ctrl_chn.protocol).error_msg(exc),
                              ctrl_chn.session)
            with suppress(Exception):
                self.mngr.shutdown(forced=True)
    @property
    def logging_id(self) -> str:
        "Returns _logging_id_ or <agent_name>[<peer.uid.hex>]"
        return getattr(self, '_logging_id_', f'{self.descriptor.agent.name}[{self.peer.uid.hex}]')



# SPDX-FileCopyrightText: 2022-present The Firebird Projects <www.firebirdsql.org>
#
# SPDX-License-Identifier: MIT
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/_scripts/commands/oids.py
# DESCRIPTION:    Saturnin OID registry management commands
# CREATED:        23.11.2022
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2022 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""Saturnin OID registry management commands

"""

from __future__ import annotations
#from typing import List
from uuid import UUID
import typer
from rich.table import Table
from rich.text import Text
from rich import box
from firebird.uuid import (oid_registry, ROOT_SPEC, get_specifications, parse_specifications,
                           Node)
from saturnin.base import directory_scheme
from saturnin.lib.console import console, _h, RICH_OK, RICH_ERROR
from saturnin._scripts.completers import oid_completer

#: Typer command group for OID management commands
app = typer.Typer(rich_markup_mode="rich", help="Saturnin OID management.")

@app.command()
def list_oids(with_name: str=typer.Option('', help="List only OIDs with this string in name"),
              show_oids: bool = typer.Option(False, '--show-oids',
                                             help="Should OIDs instead UUIDs")) -> None:
    """List registered OIDs.
    """
    table = Table(title='Registered OIDs' if not with_name
                  else f"Registered OIDs with name containing '{with_name}'",
                  box=box.ROUNDED)
    table.add_column('OID Name', style='green')
    table.add_column('OID' if show_oids else 'UUID')
    if oid_registry:
        for node in oid_registry.values():
            if with_name in node.full_name:
                if show_oids:
                    table.add_row(*[node.full_name, Text(node.oid, style='number')])
                else:
                    table.add_row(*[node.full_name, Text(str(node.uid), style='uuid')])
        console.print(table)
    else:
        console.print("No OIDs are registered.")

@app.command()
def update_oids(url: str=typer.Argument(ROOT_SPEC, help="URL to OID node specification",
                                        metavar='URL')):
    """Update OID registry from specification(s).
    """
    console.print("Downloading OID specifications ... ", end='')
    specifications, errors = get_specifications(url)
    if errors:
        console.print(RICH_ERROR)
        console.print_error("Errors occured during download:")
        for err_url, error in errors:
            console.print_error(f"URL: {err_url}")
            console.print_error(f"error: {error}")
        return
    console.print(RICH_OK)
    console.print("Parsing OID specifications ... ", end='')
    specifications, errors = parse_specifications(specifications)
    if errors:
        console.print(RICH_ERROR)
        console.print_error("Errors detected while parsing OID specifications:")
        for err_url, error in errors:
            console.print_error(f"URL: {err_url}")
            console.print_error(f"error: {error}")
        return
    console.print(RICH_OK)
    #
    console.print("Updating OID registry ... ", end='')
    try:
        oid_registry.update_from_specifications(specifications)
    except Exception as exc: # pylint: disable=W0703
        console.print(RICH_ERROR)
        console.print_error(exc)
    console.print(RICH_OK)
    directory_scheme.site_oids_toml.write_text(oid_registry.as_toml())

@app.command()
def show_oid(oid: str=typer.Argument('', help="OID name or GUID",
                                     autocompletion=oid_completer)) -> None:
    """Show information about OID.
    """
    uid: UUID = None
    name: str = None
    node: Node = None
    try:
        uid = UUID(oid)
    except Exception: # pylint: disable=W0703
        name = oid

    if uid is not None:
        node = oid_registry.get(uid)
    else:
        node = oid_registry.find(lambda x: x.full_name.startswith(name))
    if node is None:
        console.print_error('OID not registered!')
        return

    table = Table.grid()
    table.add_column(style='green')
    table.add_column()
    table.add_row('OID:', Text(node.oid, style='number'))
    table.add_row('UID:', _h(Text(str(node.uid))))
    table.add_row('Node name:', Text(node.name))
    table.add_row('Full name:', Text(node.full_name))
    table.add_row('Description:', Text(node.description))
    table.add_row('Contact:', Text(node.contact))
    table.add_row('E-mail:', _h(Text(node.email)))
    table.add_row('Site:', _h(Text(node.site)))
    table.add_row('Node spec.:', _h(Text(str(node.node_spec))))
    table.add_row('Node type:', _h(Text(node.node_type.name)))
    table.add_row('Parent spec.: ', _h(Text(str(node.parent_spec))))
    console.print(table)


if directory_scheme.site_oids_toml.is_file():
    oid_registry.update_from_toml(directory_scheme.site_oids_toml.read_text())



# SPDX-FileCopyrightText: 2020-present The Firebird Projects <www.firebirdsql.org>
#
# SPDX-License-Identifier: MIT
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/lib/data/console.py
# DESCRIPTION:    Saturnin console manager
# CREATED:        7.12.2020
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2020 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________
# pylint: disable=C0301

"""Saturnin console manager for terminal configuration and output.

"""

from __future__ import annotations
import sys
import os
from rich.console import Console
from rich.theme import Theme
from rich.text import Text
from rich.highlighter import RegexHighlighter
from saturnin.base import directory_scheme

#: Default console theme
DEFAULT_THEME: Theme = Theme(
    {'option': 'bold cyan',
     'switch': 'bold green',
     'metavar': 'bold yellow',
     'section': 'bold white',
     'help_require': 'dim',
     'args_and_cmds': 'yellow',
     'path': 'bold cyan',
     'title': 'bold yellow',
     'ok': 'green',
     'important': 'bold yellow',
     'warning': 'bold yellow',
     'error': 'bold red',
     'item': 'green',
     'attrib_equal': 'bold',
     'attrib_name': 'not italic yellow',
     'attrib_value': 'not italic magenta',
     'bool_false': 'italic bright_red',
     'bool_true': 'italic bright_green',
     'brace': 'bold',
     'call': 'bold magenta',
     'comma': 'bold',
     'ellipsis': 'yellow',
     'eui48': 'bold bright_green',
     'eui64': 'bold bright_green',
     'filename': 'bright_magenta',
     'indent': 'dim green',
     'ipv4': 'bold bright_green',
     'ipv6': 'bold bright_green',
     'py_none': 'italic magenta',
     'number': 'bold not italic cyan',
     'number_complex': 'bold not italic cyan',
     'str': 'not bold not italic green',
     'tag_contents': 'default',
     'tag_end': 'bold',
     'tag_name': 'bold bright_magenta',
     'tag_start': 'bold',
     'url': 'not bold not italic underline bright_blue',
     'zmq_address': 'not bold not italic bright_blue',
     'uuid': 'not bold bright_yellow',
     'uuid2': 'not bold bright_yellow',
     'mime': 'bold blue',
     'mime_param': 'italic cyan',
     'enum': 'bold not italic white',
     'email': 'bold magenta',
     'date': 'blue',
     'time': 'magenta',
     'timezone': 'yellow',
     })

#: Use rich terminal or not
FORCE_TERMINAL: bool = True if os.getenv("FORCE_COLOR") or os.getenv("PY_COLORS") else None

#: Standard rich text for YES
RICH_YES: Text = Text('✔', style='ok')
#: Standard rich text for NO
RICH_NO: Text = Text('✖', style='error')
#: Standard rich text for OK
RICH_OK: Text = Text('OK', style='ok')
#: Standard rich text for WARNING
RICH_WARNING: Text = Text('WARNING', style='warning')
#: Standard rich text for ERROR
RICH_ERROR: Text = Text('ERROR', style='error')

def _combine_regex(*regexes: str) -> str:
    """Combine a number of regexes in to a single regex.

    Returns:
        str: New regex with all regexes ORed together.
    """
    return "|".join(regexes)

class SaturninHighlighter(RegexHighlighter): # pylint: disable=R0903
    """Highlights our special options."""
    #: Regular expressions used by `.highlight`.
    highlights = [
        r"(^|\W)(?P<switch>\-\w+)(?![a-zA-Z0-9])",
        r"(^|\W)(?P<option>\-\-[\w\-]+)(?![a-zA-Z0-9])",
        r"(?P<metavar>\<[^\>]+\>)",
        r"(?P<section>\n[A-Z][^:\n]*(:)\s?)",
        #
        # Dates
        #
        # Calendar month (e.g. 2008-08). The hyphen is required
        r"^(?P<year>[0-9]{4})-(?P<month>1[0-2]|0[1-9])$",
        # Calendar date w/o hyphens (e.g. 20080830)
        r"^(?P<date>(?P<year>[0-9]{4})(?P<month>1[0-2]|0[1-9])(?P<day>3[01]|0[1-9]|[12][0-9]))$",
        # Ordinal date (e.g. 2008-243). The hyphen is optional
        r"^(?P<date>(?P<year>[0-9]{4})-?(?P<day>36[0-6]|3[0-5][0-9]|[12][0-9]{2}|0[1-9][0-9]|00[1-9]))$",
        #
        # Weeks
        #
        # Week of the year (e.g., 2008-W35). The hyphen is optional
        r"^(?P<date>(?P<year>[0-9]{4})-?W(?P<week>5[0-3]|[1-4][0-9]|0[1-9]))$",
        # Week date (e.g., 2008-W35-6). The hyphens are optional
        r"^(?P<date>(?P<year>[0-9]{4})-?W(?P<week>5[0-3]|[1-4][0-9]|0[1-9])-?(?P<day>[1-7]))$",
        #
        # Times
        #
        # Hours and minutes (e.g., 17:21). The colon is optional
        r"^(?P<time>(?P<hour>2[0-3]|[01][0-9]):?(?P<minute>[0-5][0-9]))$",
        # Hours, minutes, and seconds w/o colons (e.g., 172159)
        r"^(?P<time>(?P<hour>2[0-3]|[01][0-9])(?P<minute>[0-5][0-9])(?P<second>[0-5][0-9]))$",
        # Time zone designator (e.g., Z, +07 or +07:00). The colons and the minutes are optional
        r"^(?P<timezone>(Z|[+-](?:2[0-3]|[01][0-9])(?::?(?:[0-5][0-9]))?))$",
        # Hours, minutes, and seconds with time zone designator (e.g., 17:21:59+07:00).
        # All the colons are optional. The minutes in the time zone designator are also optional
        r"^(?P<time>(?P<hour>2[0-3]|[01][0-9])(?P<minute>[0-5][0-9])(?P<second>[0-5][0-9]))(?P<timezone>Z|[+-](?:2[0-3]|[01][0-9])(?::?(?:[0-5][0-9]))?)$",
        #
        # Date and Time
        #
        # Calendar date with hours, minutes, and seconds (e.g., 2008-08-30 17:21:59 or 20080830 172159).
        # A space is required between the date and the time. The hyphens and colons are optional.
        # This regex matches dates and times that specify some hyphens or colons but omit others.
        # This does not follow ISO 8601
        r"^(?P<date>(?P<year>[0-9]{4})(?P<hyphen>-)?(?P<month>1[0-2]|0[1-9])(?(hyphen)-)(?P<day>3[01]|0[1-9]|[12][0-9])) (?P<time>(?P<hour>2[0-3]|[01][0-9])(?(hyphen):)(?P<minute>[0-5][0-9])(?(hyphen):)(?P<second>[0-5][0-9]))$",
        #
        # XML Schema dates and times
        #
        # Date, with optional time zone (e.g., 2008-08-30 or 2008-08-30+07:00).
        # Hyphens are required. This is the XML Schema 'date' type
        r"^(?P<date>(?P<year>-?(?:[1-9][0-9]*)?[0-9]{4})-(?P<month>1[0-2]|0[1-9])-(?P<day>3[01]|0[1-9]|[12][0-9]))(?P<timezone>Z|[+-](?:2[0-3]|[01][0-9]):[0-5][0-9])?$",
        # Time, with optional fractional seconds and time zone (e.g., 01:45:36 or 01:45:36.123+07:00).
        # There is no limit on the number of digits for the fractional seconds. This is the XML Schema 'time' type
        r"^(?P<time>(?P<hour>2[0-3]|[01][0-9]):(?P<minute>[0-5][0-9]):(?P<second>[0-5][0-9])(?P<frac>\.[0-9]+)?)(?P<timezone>Z|[+-](?:2[0-3]|[01][0-9]):[0-5][0-9])?$",
        # Date and time, with optional fractional seconds and time zone (e.g., 2008-08-30T01:45:36 or 2008-08-30T01:45:36.123Z).
        # This is the XML Schema 'dateTime' type
        r"^(?P<date>(?P<year>-?(?:[1-9][0-9]*)?[0-9]{4})-(?P<month>1[0-2]|0[1-9])-(?P<day>3[01]|0[1-9]|[12][0-9]))T(?P<time>(?P<hour>2[0-3]|[01][0-9]):(?P<minute>[0-5][0-9]):(?P<second>[0-5][0-9])(?P<ms>\.[0-9]+)?)(?P<timezone>Z|[+-](?:2[0-3]|[01][0-9]):[0-5][0-9])?$",
        # Values
        _combine_regex(
            r"(?P<ipv4>[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3})",
            r"(?P<ipv6>([A-Fa-f0-9]{1,4}::?){1,7}[A-Fa-f0-9]{1,4})",
            r"(?P<eui64>(?:[0-9A-Fa-f]{1,2}-){7}[0-9A-Fa-f]{1,2}|(?:[0-9A-Fa-f]{1,2}:){7}[0-9A-Fa-f]{1,2}|(?:[0-9A-Fa-f]{4}\.){3}[0-9A-Fa-f]{4})",
            r"(?P<eui48>(?:[0-9A-Fa-f]{1,2}-){5}[0-9A-Fa-f]{1,2}|(?:[0-9A-Fa-f]{1,2}:){5}[0-9A-Fa-f]{1,2}|(?:[0-9A-Fa-f]{4}\.){2}[0-9A-Fa-f]{4})",
            r"(?P<uuid>[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12})",
            r"(?P<uuid2>[a-fA-F0-9]{32})",
            r"(?P<call>[\w.]*?)\(",
            r"\b(?P<bool_true>True)\b|\b(?P<bool_false>False)\b|\b(?P<py_none>None)\b",
            r"(?P<ellipsis>\.\.\.)",
            r"(?P<number_complex>(?<!\w)(?:\-?[0-9]+\.?[0-9]*(?:e[-+]?\d+?)?)(?:[-+](?:[0-9]+\.?[0-9]*(?:e[-+]?\d+)?))?j)",
            r"(?P<number>(?<!\w)\-?[0-9]+\.?[0-9]*(e[-+]?\d+?)?\b|0x[0-9a-fA-F]*)",
            r"(?P<path>\B(/[-\w._+]+)*\/)(?P<filename>[-\w._+]*)?",
            r"(?<![\\\w])(?P<str>b?'''.*?(?<!\\)'''|b?'.*?(?<!\\)'|b?\"\"\".*?(?<!\\)\"\"\"|b?\".*?(?<!\\)\")",
            r"(?P<url>(file|https|http|ws|wss)://[-0-9a-zA-Z$_+!`(),.?/;:&=%#]*)",
            r"(?P<zmq_address>(inproc|ipc|tcp|pgm|epgm|vmci)://[-0-9a-zA-Z$_+!`(),.?/;:&=%#]*)",
            r"(?P<mime>(application|audio|font|example|image|message|model|multipart|text|video|x-(?:[0-9A-Za-z!#$%&'*+.^_`|~-]+))/([0-9A-Za-z!#$%&'*+.^_`|~-]+))(?P<mime_param>((?:[ \t]*;[ \t]*[0-9A-Za-z!#$%&'*+.^_`|~-]+=(?:[0-9A-Za-z!#$%&'*+.^_`|~-]+|\"(?:[^\"\\\\]|\\.)*\"))*))",
            r"(?P<enum>\b[A-Z_]*\b)",
            r"(?P<email>[.\w-]+@([\w-]+\.)+[\w-]+)",
        )
    ]
    def highlight(self, text: Text) -> Text:
        """Highlight `~rich.text.Text` using regular expressions.

        Arguments:
            text: Text to highlighted.
        """
        highlight_regex = text.highlight_regex
        for re_highlight in self.highlights:
            highlight_regex(re_highlight, style_prefix=self.base_style)
        return text

#: Saturnin text highlighter
highlighter: SaturninHighlighter = SaturninHighlighter()
#: Shortcut to `highlighter.highlight`
_h = highlighter.highlight

class ConsoleManager():
    """Saturnin site manager.
    """
    def __init__(self):
        #: Suppress output flag
        self.quiet: bool = False
        #: Verbose output flag
        self.verbose: bool = False
        #: Rich main console
        self.std_console: Console = Console(theme=Theme.read(directory_scheme.theme_file)
                                            if directory_scheme.theme_file.exists()
                                            else DEFAULT_THEME, tab_size=4, #emoji=False,
                                            highlighter=highlighter, highlight=True,
                                            force_terminal=FORCE_TERMINAL)
        if not sys.stdout.isatty():
            self.std_console.width = 5000
        #: Rich error console
        self.err_console: Console = Console(stderr=True, style='bold red', tab_size=4, #emoji=False,
                                            force_terminal=FORCE_TERMINAL)
    def print_info(self, message='') -> None:
        "Prints information message to console."
        if self.verbose and not self.quiet:
            if message:
                self.std_console.print(message, style='yellow')
            else:
                self.std_console.print()
    def print_error(self, message) -> None:
        "Prints error message to error console."
        self.err_console.print(message)
    def print_exception(self) -> None:
        "Prints exception to error console."
        self.err_console.print_exception()
    def print(self, message = '', end='\n') -> None:
        "Prints message to console."
        if not self.quiet:
            self.std_console.print(message, end=end, highlight=True)

#: Saturnin site manager
console: ConsoleManager = ConsoleManager()



# SPDX-FileCopyrightText: 2021-present The Firebird Projects <www.firebirdsql.org>
#
# SPDX-License-Identifier: MIT
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/_scripts/cli.py
# DESCRIPTION:    Script for Saturnin CLI manager
# CREATED:        19.1.2021
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2021 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________
# pylint: disable=W0212, C0301, W0703, W0622

"""Saturnin CLI manager.

---

\b
When invoked without additional parameters, it's activated in **console** mode.
This mode runs REPL that provides advanced functionality like command and parameter completion,
interactive help and persistent command history. The command set available in this mode
may differ from command set available in **direct** mode (see below).

When saturnin is invoked with additional parameters, it executes specified command and exists.
Some commands (typically those required to run only once or not very often like *initialize*
or *create home*) are available only in this **direct** mode.
"""

# group commands
# LIST - print list of items (services, nodes, packages etc.)
# SHOW - print details about particular item
# EDIT - Edit items (configuration etc.)
# CREATE - create item
# RUN - run components
# START - ??
# STOP - Stop activities (daemons, notifications, trace, logging)
# INSTALL - install components via pip
# UNINSTALL - uninstall components via pip
# UPDATE - Update items (components, OIDs etc.)

from __future__ import annotations
from typing import Callable, List, Tuple, Optional
import sys
from typer import Typer, Context
from rich.align import Align
from rich.padding import Padding
from rich.markdown import Markdown
from saturnin.component.recipe import recipe_registry
from saturnin.component.apps import application_registry
from saturnin.component.registry import iter_entry_points
from saturnin.lib.console import console
# from saturnin.lib import wingdbstub
from .repl import repl, IOManager
from .commands.recipes import run_recipe
from .completers import get_first_line

#: REPL introductory markdown text
REPL_INTRO = Markdown("""
# Saturnin management console

Saturnin was invoked in **console** mode that provides command and parameter completion,
interactive help and persistent command history. The command set available in this mode may
differ from command set available in **direct** mode (see below).

When saturnin is invoked with additional parameters, it executes specified command and exists.
Some commands (typically those required to run only once or not very often like *initialize*
or *create home*) are available only in this **direct** mode.

---
""")

#: REPL help text
REPL_HELP = """Type '?' or '?<command>' for help.
Type 'quit' to leave the console.
Use 'Ctrl+Space' to activate command completion.
"""

#: Standard command groups
#: List of (command name, short help) tuples
command_groups: List[Tuple[str,str]] = [
    ('list', "Print list of items (services, nodes, packages etc.)"),
    ('show', "Print details about particular item (service, node, package etc.)"),
    ('edit', "Edit item (configuration, recipe etc.)"),
    ('create', "Create item (configuration, recipe etc.)"),
    ('run', "Run Saturnin components (recipes, applications, utilities etc.)"),
    ('update', "Update items (components, OIDs etc.)"),
    ('stop', "Stop activities (daemons, notifications, trace, logging)"),
    ('install', "Install components (packages, recipes etc.)"),
    ('uninstall', "Uninstall components (packages, recipes etc.)"),
]


def find_group(in_app: Typer, name: str) -> Typer:
    """Returns sub-command group in command group.

    Arguments:
      in_app: Typer instance to be searched.
      name:   Command name.

    Returns: Typer instance for command.
    """
    for grp in in_app.registered_groups:
        if grp.name == name:
            return grp.typer_instance
    return None

def add_command(app: Typer, name: str, cmd: Callable, *, help: Optional[str]=None,
                panel: Optional[str]=None) -> None: # pylint: disable=W0622
    """Add command into main Typer application.

    Arguments:
       app:   Typer instance under which the command should be placed.
       name:  Command name. Can use dot notation to create a sub-command.
       cmd:   Callable to be registered as Typer command.
       help:  Optional help for command.
       pabel: Rich panel where command should be listed in help.
    """
    names: List[str] = name.split('.')
    group: Typer = app
    for group_name in names[:-1]:
        sub_group = find_group(group, group_name)
        if sub_group is None:
            sub_group = Typer(name=group_name)
            group.add_typer(sub_group)
            group = sub_group
        else:
            group = sub_group
    group.command(name=names[-1], help=help, rich_help_panel=panel)(cmd)

def cli_loop(*, restart: bool) -> bool:
    """Main CLI loop via Typer.

    Arguments:
      restart: True when CLI was restarted.

    Returns:
      True if CLI restart is required (to reload commands etc.). False means "normal" end.
    """
    def replcheck(ctx: Context):
        """This method start REPL when CLI is started without subcommand.
        """
        if not app._in_repl and ctx.invoked_subcommand is None:
            app._in_repl = True
            if not app._restart:
                console.print(Padding(Align(REPL_INTRO, pad=False,),(1, 1, 1, 1),))
                console.print(Padding(Align(REPL_HELP, pad=False,),(0, 1, 0, 1),))
            else:
                ctx.command.help = ''
            with IOManager(ctx) as ioman:
                app._restart = repl(ctx, ioman)

    app: Typer = None
    going_repl = len(sys.argv) == 1
    if going_repl:
        app = Typer(rich_markup_mode="markdown", epilog=REPL_HELP)
    else:
        app = Typer(rich_markup_mode="markdown", help=__doc__)
    app._in_repl = False
    app._restart = restart
    app.callback(invoke_without_command=True)(replcheck)
    # Install command groups
    for group_name, group_help in command_groups:
        app.add_typer(Typer(), name=group_name, help=group_help)
    # Install registered commands
    for entry in iter_entry_points('saturnin.commands'):
        try:
            add_command(app, entry.name, entry.load())
        except Exception as exc:
            console.print_error(f"Cannot install command '{entry.name}'\n{exc!s}")
    if going_repl:
        for entry in iter_entry_points('saturnin.repl_only_commands'):
            try:
                add_command(app, entry.name, entry.load())
            except Exception as exc:
                console.print_error(f"Cannot install command '{entry.name}'\n{exc!s}")
    else:
        for entry in iter_entry_points('saturnin.no_repl_commands'):
            try:
                add_command(app, entry.name, entry.load())
            except Exception as exc:
                console.print_error(f"Cannot install command '{entry.name}'\n{exc!s}")
    # Install registered recipes
    for recipe in recipe_registry.values():
        if recipe.application is None:
            try:
                add_command(app, f'run.{recipe.name}', run_recipe,
                            help=recipe.description)
            except Exception as exc:
                console.print_error(f"Cannot install command '{recipe.name}'\n{exc!s}")
        else:
            application = application_registry.get(recipe.application)
            if application is None:
                console.print_error(f"Cannot install recipe {recipe.name} due to missing application.")
                continue
            try:
                add_command(app, f'run.{recipe.name}', application.factory_obj,
                            help=get_first_line(recipe.description))
            except Exception as exc:
                console.print_error(f"Cannot install command '{recipe.name}'\n{exc!s}")
    #
    app(standalone_mode=False)
    return app._restart

def main():
    """Main entry point for Saturnin CLI manager.
    """
    restart = False
    while restart := cli_loop(restart=restart):
        pass

if __name__ == '__main__':
    main()



# SPDX-FileCopyrightText: 2022-present The Firebird Projects <www.firebirdsql.org>
#
# SPDX-License-Identifier: MIT
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/_scripts/commands/recipes.py
# DESCRIPTION:    Saturnin recipe commands
# CREATED:        29.11.2022
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2023 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________
# pylint: disable=R0912, R0913, R0914, R0915

"""Saturnin recipe commands


"""

from __future__ import annotations
from typing import List, Tuple
from pathlib import Path
from tempfile import TemporaryDirectory
from configparser import ConfigParser, ExtendedInterpolation
from uuid import UUID
from datetime import datetime
import subprocess
import typer
from rich.table import Table
from rich.text import Text
from rich.markdown import Markdown
from rich.syntax import Syntax
from rich import box
from firebird.uuid import oid_registry, Node
from firebird.base.config import Config
from saturnin.base import (saturnin_config, SECTION_BUNDLE, SECTION_SERVICE,
                           directory_scheme, RESTART)
from saturnin.component.recipe import (recipe_registry, RecipeInfo, SaturninRecipe,
                                       RecipeType, RecipeExecutionMode)
from saturnin.component.registry import service_registry, ServiceInfo
from saturnin.lib.console import console, _h, RICH_YES, RICH_NO
from saturnin.component.apps import application_registry, ApplicationInfo
from saturnin.component.bundle import ServiceBundleConfig
from saturnin.component.controller import ServiceExecConfig
from saturnin._scripts.completers import (recipe_completer, service_completer,
                                          path_completer, application_completer,
                                          get_first_line)

#: Typer command group for recipe management commands
app = typer.Typer(rich_markup_mode="markdown", help="Saturnin recipes.")

def run_recipe(ctx: typer.Context,
               section: str=typer.Option(None, help="Main recipe section name"),
               print_outcome: bool=typer.Option(False, '--print-outcome',
                                                help="Print service execution outcome"),
               config: List[str]=typer.Option(None,
                                              help="Path to additional configuration file "
                                              "(could be specified multiple times)"),
               quiet: bool=typer.Option(False, '--quiet', help="Suppress console output"),
               main_thread: bool=typer.Option(False, '--main-thread',
                                              help="Start the service in main thread. "
                                              "Ignored for bundles.")) -> None:
    """Runs Saturnin recipe.
    """
    recipe_name = ctx.command.name
    recipe: RecipeInfo = recipe_registry.get(recipe_name)
    if recipe is None:
        console.print_error(f"Recipe '{recipe_name}' not installed")
        return
    if recipe.recipe_type is RecipeType.BUNDLE:
        cmd = ['saturnin-bundle' if recipe.executor is None else str(recipe.executor)]
    else:
        cmd = ['saturnin-service' if recipe.executor is None else str(recipe.executor)]
        if main_thread:
            cmd.append('--main-thread')
    if section:
        cmd.extend(['--section', section])
    if quiet:
        cmd.append('--quiet')
    if print_outcome:
        cmd.append('--outcome')
    for cfg in config:
        cmd.extend(['-c', str(cfg)])
    cmd.append(str(recipe.filename))
    # Daemonize
    pid_file: Path = None
    with TemporaryDirectory() as tmp_dir:
        if recipe.execution_mode is RecipeExecutionMode.DAEMON:
            cmd.insert(1, '-q')
            pid_file = Path(tmp_dir) / 'pid'
            cmd.insert(0, str(pid_file))
            cmd.insert(0, '-p')
            cmd.insert(0, 'start')
            cmd.insert(0, 'saturnin-daemon')
        start = datetime.now()
        result = subprocess.run(cmd) # pylint: disable=W1510
        console.print(f'Execution time: {datetime.now() - start}')
        if pid_file:
            if pid_file.exists():
                pid = int(pid_file.read_text())
                pid_file = directory_scheme.pids / f'{pid}.pid'
                pid_file.write_text(recipe_name)
            else:
                console.print_error("Daemonized recipe execution failed [PID file not found]")
    #result = subprocess.run(cmd, capture_output=True, text=True)
    #console.print(result.stdout)
    if result.returncode != 0:
        console.print_error('Recipe execution failed')

@app.command()
def list_recipes() -> None:
    """List installed Saturnin recipes.
    """
    if recipe_registry:
        table = Table(title='Installed recipes', box=box.ROUNDED)
        table.add_column('Name', style='green')
        table.add_column('Type', style='enum')
        table.add_column('Execution mode', style='enum')
        table.add_column('App', width=3, justify='center')
        table.add_column('Description')
        recipe: RecipeInfo = None
        for recipe in recipe_registry.values():
            table.add_row(recipe.name, recipe.recipe_type.name, recipe.execution_mode.name,
                          RICH_NO if recipe.application is None else RICH_YES,
                          get_first_line(recipe.description))
        console.print(table)
    else:
        console.print("There are no Saturnin recipes installed.")

@app.command()
def show_recipe(recipe_name: str=typer.Argument(..., help="Recipe name",
                                                autocompletion=recipe_completer),
                section: str=typer.Option(None, help="Configuration section name"),
                raw: bool=typer.Option(False, '--raw',
                                       help="Print recipe file content instead normal output")):
    """It analyzes the content of the recipe and displays its structure and configuration according to the default
   sections of the container configuration. If the recipe contains several variants, it is necessary to enter
   the name of the specific section for the configuration of the container to display them.

   Alternatively, it is possible to display the entire recipe in text form (with syntax highlighting).

    """
    recipe: RecipeInfo = recipe_registry.get(recipe_name)
    if recipe is None:
        console.print_error(f"Recipe '{recipe_name}' not installed")
        return
    if raw:
        console.print(Syntax(recipe.filename.read_text(), 'cfg',word_wrap=True))
    else:
        config: ConfigParser = ConfigParser(interpolation=ExtendedInterpolation())
        config.read(recipe.filename)
        recipe_config = SaturninRecipe()
        recipe_config.load_config(config)
        recipe_config.validate()
        #
        title = "default section" if section is None else f'section "[section]{section}[/]"'
        services: List[Tuple[str, ServiceInfo]] = []
        if section is None:
            section = SECTION_BUNDLE if recipe_config.recipe_type.value is RecipeType.BUNDLE \
                else SECTION_SERVICE
        if config.has_section(section):
            if recipe_config.recipe_type.value is RecipeType.BUNDLE:
                bundle_cfg: ServiceBundleConfig = ServiceBundleConfig(section)
                bundle_cfg.load_config(config)
                for agent_config in bundle_cfg.agents.value:
                    services.append((agent_config.name, service_registry[agent_config.agent.value]))
            else:
                svc_cfg: ServiceExecConfig = ServiceExecConfig(section)
                svc_cfg.load_config(config)
                services.append(('service', service_registry[svc_cfg.agent.value]))
        #
        table = Table.grid()
        table.add_column('', style='green')
        table.add_column('')
        table.add_row(' Name:', Text(recipe.name, style='bold cyan'))
        table.add_row(' Type:', Text(recipe.recipe_type.name, style='enum'))
        table.add_row(' Exec. mode:', Text(recipe.execution_mode.name, style='enum'))
        table.add_row(' Executor:', _h(Text('DEFAULT' if recipe.executor is None
                                            else str(recipe.executor))))
        table.add_row(' Application:', _h(Text(str(recipe.application) if recipe.application
                                               else '')))
        table.add_row(' Description: ', Markdown(recipe.description))
        console.print(table)
        console.print()
        if not services:
            if recipe_config.application.has_value():
                console.print("[important] Recipe components deternimed dynamically by application.")
            else:
                console.print_error(f"Recipe does not have expected section {section}")
        else:
            table = Table(title=f"  Recipe components ({title})", box=box.ROUNDED, title_justify='left')
            table.add_column('Cfg. name', style='bold yellow')
            table.add_column('Component', style='green')
            table.add_column('Version', style='number')
            table.add_column('Description')
            for cfg_name, svc in services:
                table.add_row(cfg_name, svc.name, svc.version, get_first_line(svc.description))
            console.print(table)
            console.print()
            for cfg_name, svc in services:
                table = Table(title=f"  [title]{cfg_name}[/] configuration", box=box.ROUNDED,
                              title_justify='left', highlight=True)
                table.add_column('Parameter', style='green')
                table.add_column('Value')
                cmp_conf: Config = svc.descriptor_obj.config()
                cmp_conf.load_config(config, cfg_name)
                for option in cmp_conf.options:
                    table.add_row(option.name, 'None' if option.get_value() is None
                                  else option.get_as_str())
                console.print(table)
                console.print()

@app.command()
def edit_recipe(recipe_name: str=typer.Argument(..., help="Recipe name",
                                                autocompletion=recipe_completer)):
    """Edit recipe.
    """
    recipe: RecipeInfo = recipe_registry.get(recipe_name)
    if recipe is None:
        console.print_error(f"Recipe '{recipe_name}' not installed")
        return None
    edited = typer.edit(recipe.filename.read_text(), saturnin_config.editor.value,
                        extension=recipe.filename.suffix)
    if edited is not None:
        recipe.filename.write_text(edited)
        console.print("Recipe updated.")
    recipe_registry.clear()
    recipe_registry.load_from(directory_scheme.recipes)
    return RESTART

@app.command()
def install_recipe(recipe_name: str= \
                     typer.Option(None, help="Recipe name (default is recipe file name / application name)"),
                   recipe_file: Path= \
                     typer.Option(None, help="Recipe file. Mutually exclusive with ",
                                  dir_okay=False, autocompletion=path_completer),
                   app_id: str=typer.Option(None, help="Application UID or name",
                                            autocompletion=application_completer)):
    """Installs a new recipe from an external recipe file or from an installed application.
    Once installed, recipe can be executed immediately with the `run <recipe-name>` command.
    """
    if recipe_file is None and app_id is None:
        console.print_error("Either recipe file or application must be specified.")
        return None
    if recipe_file is not None and app_id is not None:
        console.print_error("Either recipe file or application must be specified, but not both")
        return None
    app: ApplicationInfo = None
    if app_id is not None:
        try:
            app = application_registry.get(UUID(app_id))
        except Exception: # pylint: disable=W0703
            app = application_registry.get_by_name(app_id)
        if app is None:
            console.print_error('Application not registered!')
            return None
    if recipe_name is None:
        if app is None:
            recipe_name = recipe_file.stem
        else:
            recipe_name = app.get_recipe_name()
    #
    if recipe_name in recipe_registry:
        console.print_error(f"Recipe '{recipe_name}' already installed")
        return None
    #
    target: Path = directory_scheme.recipes
    config: ConfigParser = ConfigParser(interpolation=ExtendedInterpolation())
    if app is None: # Import recipe
        config.read(recipe_file)
        recipe_content = recipe_file.read_text()
        target = target / recipe_file.with_stem(recipe_name).name
    else: # Import application
        recipe_content = app.config_obj()
        config.read_string(recipe_content)
        target = target / f'{recipe_name}.cfg'

    recipe_config = SaturninRecipe()
    recipe_config.load_config(config)
    recipe_config.validate()
    # Check whether all required components are installed
    application = recipe_config.application.value
    if (application and application not in application_registry):
        node: Node = oid_registry.get(application)
        console.print_error(f"Required application '{node.name if node else application}' not installed")
        return None
    section = SECTION_BUNDLE if recipe_config.recipe_type.value is RecipeType.BUNDLE \
        else SECTION_SERVICE
    services: List[UUID] = []
    if config.has_section(section):
        if recipe_config.recipe_type.value is RecipeType.BUNDLE:
            bundle_cfg: ServiceBundleConfig = ServiceBundleConfig(section)
            bundle_cfg.load_config(config)
            for agent_config in bundle_cfg.agents.value:
                services.append(agent_config.agent.value)
        else:
            svc_cfg: ServiceExecConfig = ServiceExecConfig(section)
            svc_cfg.load_config(config)
            services.append(svc_cfg.agent.value)
        stop = False
        for svc_uid in services:
            if svc_uid not in service_registry:
                node: Node = oid_registry.get(svc_uid)
                console.print_error(f"Required service '{node.name if node else svc_uid}' not installed")
                stop = True
        if stop:
            return None
    else:
        console.print_error(f"The recipe does not have the required section '{section}'")
        return None
    #
    target.write_text(recipe_content)
    console.print("Recipe installed.")
    recipe_registry.clear()
    recipe_registry.load_from(directory_scheme.recipes)
    return RESTART

@app.command()
def uninstall_recipe(recipe_name: str=typer.Argument(None, autocompletion=recipe_completer,
                                                     help="The name of the recipe to be uninstalled"),
                     save_to: Path=typer.Option(None, dir_okay=False, writable=True,
                                                help="File where recipe should be saved before it's removed")):
    """Uninstall recipe. Can optionally save the recipe file
    """
    recipe: RecipeInfo = recipe_registry.get(recipe_name)
    if recipe is None:
        console.print_error(f"Recipe '{recipe_name}' not installed")
        return None
    if save_to is not None:
        save_to.write_text(recipe.filename.read_text())
    recipe.filename.unlink()
    console.print("Recipe uninstalled.")
    recipe_registry.clear()
    recipe_registry.load_from(directory_scheme.recipes)
    return RESTART

@app.command()
def create_recipe(plain: bool=typer.Option(False, '--plain', help="Create recipe without comments"),
                  recipe_name: str=typer.Argument(..., help="Recipe name", metavar='NAME'),
                  components: List[str]=typer.Argument(..., help="Recipe components",
                                                       autocompletion=service_completer)):
    """Creates a recipe template that uses the specified Butler services. Such a template
    contains only default settings and usually needs to be modified to achieve the desired
    results.
    """
    if recipe_name in recipe_registry:
        console.print_error(f"Recipe '{recipe_name}' already exists")
        return None
    #
    recipe_type: RecipeType = RecipeType.SERVICE if len(components) == 1 \
        else RecipeType.BUNDLE
    service_config: ServiceExecConfig = ServiceExecConfig(SECTION_SERVICE)
    bundle_config: ServiceBundleConfig = ServiceBundleConfig(SECTION_BUNDLE)
    recipe_config = SaturninRecipe()
    recipe_config.recipe_type.value = recipe_type
    #
    for i, component in enumerate(components, 1):
        svc: ServiceInfo = None
        try:
            svc = service_registry.get(UUID(component))
        except Exception: # pylint: disable=W0703
            svc = service_registry.get_by_name(component)
        if svc is None:
            console.print_error('Service not registered!')
            return None
        service_config.agent.value = svc.uid
        component_config = svc.descriptor_obj.config()
        if recipe_type is RecipeType.BUNDLE:
            component_config._name = f'component_{i}'
            service_config._name = f'component_{i}'
            bundle_config.agents.value.append(component_config)
        else:
            component_config._name = SECTION_SERVICE
    #
    recipe_text: str = recipe_config.get_config(plain=plain)
    recipe_text += '\n'
    recipe_text += bundle_config.get_config(plain=plain) if recipe_type is RecipeType.BUNDLE\
        else component_config.get_config(plain=plain)
    target: Path = directory_scheme.recipes / f'{recipe_name}.cfg'
    recipe_text = typer.edit(recipe_text, saturnin_config.editor.value,
                             extension='.cfg', require_save=False)
    if recipe_text is not None:
        target.write_text(recipe_text)
        console.print("Recipe created.")
        recipe_registry.clear()
        recipe_registry.load_from(directory_scheme.recipes)
        return RESTART
    return None



# SPDX-FileCopyrightText: 2021-present The Firebird Projects <www.firebirdsql.org>
#
# SPDX-License-Identifier: MIT
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/_scripts/commands/site.py
# DESCRIPTION:    Saturnin site manager commands
# CREATED:        11.3.2021
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2021 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""Saturnin site manager commands


"""

from __future__ import annotations
from pathlib import Path
from enum import Enum
import typer
from rich.panel import Panel
from rich.table import Table
from rich import box
from rich.prompt import Confirm
from rich.syntax import Syntax
from saturnin.base import CONFIG_HDR, directory_scheme, saturnin_config, venv
from saturnin.lib.console import console, DEFAULT_THEME, RICH_YES, RICH_NO, RICH_OK
from saturnin._scripts.completers import path_completer

#: Typer command group for site management commands
app = typer.Typer(rich_markup_mode="rich", help="Saturnin site management.")

def ensure_dir(description: str, path: Path):
    """Create directory (incl. parents) if it does not exists.

    Arguments:
      description: Directory description.
      path: Directory path.
    """
    console.print(f"{description}: [path]{path}[/path] ... ", end='')
    if not path.exists():
        path.mkdir(parents=True)
    console.print(RICH_OK)

def ensure_config(path: Path, content: str, new_config: bool):
    """Create configuration file if it does not exists.

    Arguments:
      path: Configuration file.
      content: Content to be written into configuration file.
      new_config: When True, the configuration file is written even if it elready exists,
                  but original file is kept as renamed with '.bak' suffix.
    """
    if path.is_file():
        if not new_config:
            console.print(f"  Info : [path]{path}[/path] already exists.")
            return
        path.replace(path.with_suffix(path.suffix + '.bak'))
    console.print(f"  Writing : [path]{path}[/path] ... ", end='')
    path.write_text(content)
    console.print(RICH_OK)

def add_path(table: Table, description: str, path: Path) -> None:
    """Adds new row to table with information about path incl. indicator whether path exists.

    Arguments:
      table: Rich table
      description: Path description
      path: Path
    """
    table.add_row(description, RICH_YES if path.exists() else RICH_NO, str(path))


@app.command()
def create_home() -> None:
    """Creates Saturnin home directory in Saturnin virtual environment.

    ---

    #### Important:

    To have desired effect, this command must be executed BEFORE **initialize**.
    """
    ensure_dir('Saturnin HOME', venv() / 'home')

@app.command()
def initialize(new_config: bool= \
                 typer.Option(False, '--new-config',
                              help="Create configuration files even if they already exist."),
               yes: bool= \
                 typer.Option(False, '--yes',
                              help="Don’t ask for confirmation of site initialization.")) -> None:
    """Initialize Saturnin environment/installation.

    ---

    It creates required directories and configuration files.

    #### Important:

    Before you execute this command, consider where you want to place numerous Saturnin
    directories. By default, Saturnin uses directory scheme according to platform standards.

    * If you want to place main directories under central home directory, set **SATURNIN_HOME**
      environment variable to point to this directory.

    * If you want to place home directory in Saturnin virtual environment, first execute
      the **create home** command.

    """
    if not yes:
        yes = Confirm.ask("Are you sure you want to initialize the Saturnin environment?")
    if yes:
        saturnin_cfg = CONFIG_HDR + saturnin_config.get_config()
        steps = [(console.print, ['Ensuring existence of Saturnin directories...']),
                 (ensure_dir, ["  Saturnin configuration      ", directory_scheme.config]),
                 (ensure_dir, ["  Saturnin data               ", directory_scheme.data]) ,
                 (ensure_dir, ["  Run-time data               ", directory_scheme.run_data]) ,
                 (ensure_dir, ["  Log files                   ", directory_scheme.logs]) ,
                 (ensure_dir, ["  Temporary files             ", directory_scheme.tmp]) ,
                 (ensure_dir, ["  Cache                       ", directory_scheme.cache]) ,
                 (ensure_dir, ["  User-specific configuration ", directory_scheme.user_config]) ,
                 (ensure_dir, ["  User-specific data          ", directory_scheme.user_data]) ,
                 (ensure_dir, ["  PID files                   ", directory_scheme.pids]) ,
                 (ensure_dir, ["  Recipes                     ", directory_scheme.recipes]) ,
                 (console.print, ['Creating configuration files...']),
                 (ensure_config, [directory_scheme.site_conf, saturnin_cfg, new_config]) ,
                 (ensure_config, [directory_scheme.user_conf, saturnin_cfg, new_config]) ,
                 (ensure_config, [directory_scheme.theme_file, DEFAULT_THEME.config, new_config]) ,
                 ]
        for func, params in steps:
            func(*params)

@app.command()
def list_directories() -> None:
    """List Saturnin directories.
    Emojis: :heavy_check_mark: :heavy_multiplication_x:
    """
    tbl = Table.grid(padding=(0, 1, 0, 1))
    tbl.add_column(style='white')
    tbl.add_column(width=1)
    tbl.add_column(style='path')
    if directory_scheme.has_home_env():
        tbl.add_row('[bold]SATURNIN_HOME[/bold] is set to', ':', str(directory_scheme.home))
    else:
        tbl.add_row('[important]SATURNIN_HOME env. variable not defined', '', '')
    add_path(tbl, 'Saturnin configuration', directory_scheme.config)
    add_path(tbl, 'Saturnin data', directory_scheme.data)
    add_path(tbl, 'Run-time data', directory_scheme.run_data)
    add_path(tbl, 'Log files', directory_scheme.logs)
    add_path(tbl, 'Temporary files', directory_scheme.tmp)
    add_path(tbl, 'Cache', directory_scheme.cache)
    add_path(tbl, 'User-specific configuration', directory_scheme.user_config)
    add_path(tbl, 'User-specific data', directory_scheme.user_data)
    add_path(tbl, 'PID files', directory_scheme.pids)
    add_path(tbl, 'Recipes', directory_scheme.recipes)
    console.print(Panel(tbl, title='[title]Saturnin directories',
                        title_align='left', box=box.ROUNDED))

@app.command()
def list_configs() -> None:
    """List Saturnin configuration files.
    """
    tbl = Table.grid(padding=(0, 1, 0, 1))
    tbl.add_column(style='white')
    tbl.add_column(width=1)
    tbl.add_column(style='path')
    add_path(tbl, 'Main configuration', directory_scheme.site_conf)
    add_path(tbl, 'User configuration', directory_scheme.user_conf)
    add_path(tbl, 'Console theme', directory_scheme.theme_file)
    add_path(tbl, 'Firebird configuration', directory_scheme.firebird_conf)
    add_path(tbl, 'Logging configuration', directory_scheme.logging_conf)
    console.print(Panel(tbl, title='[title]Configuration files', title_align='left',
                        box=box.ROUNDED))

@app.command()
def list_datafiles() -> None:
    """List Saturnin data files.
    """
    tbl = Table.grid(padding=(0, 1, 0, 1))
    tbl.add_column(style='white')
    tbl.add_column(width=1)
    tbl.add_column(style='path')
    add_path(tbl, 'Installed services', directory_scheme.site_services_toml)
    add_path(tbl, 'Installed applications', directory_scheme.site_apps_toml)
    add_path(tbl, 'Registered OIDs', directory_scheme.site_oids_toml)
    add_path(tbl, 'Console history', directory_scheme.history_file)
    add_path(tbl, 'Default log file', directory_scheme.log_file)
    console.print(Panel(tbl, title='[title]Saturnin data files', title_align='left',
                        box=box.ROUNDED))

class _Configs(Enum):
    "Saturnin configuration files"
    MAIN = 'main'
    USER = 'user'
    FIREBIRD = 'firebird'
    LOGGING = 'logging'
    THEME = 'theme'
    @property
    def path(self) -> Path:
        "Path to configuration file."
        if self is _Configs.MAIN:
            return directory_scheme.site_conf
        if self is _Configs.USER:
            return directory_scheme.user_conf
        if self is _Configs.FIREBIRD:
            return directory_scheme.firebird_conf
        if self is _Configs.LOGGING:
            return directory_scheme.logging_conf
        return directory_scheme.theme_file


@app.command()
def show_config(config_file: _Configs = typer.Argument(..., help="Configuration file")):
    """Show content of configuration file.
    """
    lexer = config_file.path.suffix[1:]
    if lexer == 'conf':
        lexer = 'cfg'
    console.print(Syntax(config_file.path.read_text(), lexer))

@app.command()
def edit_config(config_file: _Configs = typer.Argument(..., help="Configuration file",
                                                   autocompletion=path_completer)):
    """Edit configuration file.
    """
    edited = typer.edit(config_file.path.read_text(), saturnin_config.editor.value,
                        extension=config_file.path.suffix)
    if edited is not None:
        config_file.path.write_text(edited)
        console.print("Configuration updated.")

@app.command()
def create_config(config_file: _Configs= \
                    typer.Argument(..., help="Configuration file to be created"),
                  new_config: bool= \
                    typer.Option(False, '--new-config',
                                 help="Create configuration file even if it already exist.")):
    """Creates configuration file with default content.
    """
    config: str = None
    if config_file in (_Configs.MAIN, _Configs.USER):
        config = CONFIG_HDR + saturnin_config.get_config()
    elif config_file is _Configs.THEME:
        config = DEFAULT_THEME.config
    elif config_file is _Configs.FIREBIRD:
        try:
            from firebird.driver import driver_config # pylint: disable=C0415
            srv_cfg = """[local]
host = localhost
user = SYSDBA
password = masterkey
"""
            driver_config.register_server('local', srv_cfg)
            config = driver_config.get_config()
        except Exception: # pylint: disable=W0703
            console.print_error("Firebird driver not installed.")
            return None
    elif config_file is _Configs.LOGGING:
        config = f"""
; =====================
; Logging configuration
; =====================
;
; For details see https://docs.python.org/3/howto/logging.html#configuring-logging and
; https://docs.python.org/3/library/logging.config.html#logging-config-fileformat

[loggers]
keys = root, saturnin, trace

[handlers]
keys = file, trace, stderr

[formatters]
keys = simple, context

[logger_root]
handlers = stderr

[logger_saturnin]
handlers = file
qualname=saturnin

[logger_trace]
handlers = trace
qualname=trace
level=DEBUG
propagate=0

[handler_file]
; This handler sends logging output to file in log directory
class = FileHandler
args = ('{directory_scheme.log_file}', 'w')
formatter=context

[handler_stderr]
; This handler sends logging output to STDERR
class = StreamHandler
args = (sys.stderr,)
formatter=simple

[handler_trace]
; This handler sends logging output to STDERR
class = StreamHandler
args = (sys.stderr,)
formatter=context

[formatter_simple]
format = %(asctime)s %(levelname)s [%(process)s/%(thread)s] %(message)s

[formatter_context]
format = %(asctime)s %(levelname)s [%(processName)s/%(threadName)s] [%(agent)s:%(context)s] %(message)s
"""
    ensure_config(config_file.path, config, new_config)
    return None



# SPDX-FileCopyrightText: 2021-present The Firebird Projects <www.firebirdsql.org>
#
# SPDX-License-Identifier: MIT
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/_scripts/commands/pkg.py
# DESCRIPTION:    Saturnin package manager commands
# CREATED:        19.1.2021
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2021 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________
 # pylint: disable=W1510

"""Saturnin package manager commands


"""

from __future__ import annotations
from typing import List
import subprocess
from uuid import UUID
from re import sub
from operator import attrgetter
from rich.table import Table
from rich.text import Text
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich import box
import typer
from firebird.uuid import oid_registry
from saturnin.base import directory_scheme, RESTART
from saturnin.component.recipe import recipe_registry
from saturnin.component.registry import service_registry, ServiceInfo
from saturnin.component.apps import application_registry, ApplicationInfo
from saturnin.lib.console import console, _h, RICH_YES, RICH_NO
from saturnin.lib.metadata import distribution
from saturnin._scripts.completers import service_completer, application_completer, get_first_line

#: Typer command group for package management commands
app = typer.Typer(rich_markup_mode="rich", help="Package management.")

@app.command()
def list_services(with_name: str= \
                    typer.Option('',help="List only services with this string in name")):
    "Lists installed Saturnin services."
    services = list(service_registry.filter(lambda x: with_name in x.name))
    services.sort(key=attrgetter('name'))
    if services:
        table = Table(title='Registered services' if not with_name
                      else f"Registered services with name containing '{with_name}'",
                      box=box.ROUNDED)
        table.add_column('Service', style='green')
        table.add_column('Version', style='number')
        table.add_column('Description')
        for svc in services:
            table.add_row(svc.name, svc.version, get_first_line(svc.description))
        console.print(table)
    else:
        console.print("There are no Saturnin services registered.")

@app.command()
def show_service(service_id: str=typer.Argument('', help="Service UID or name",
                                                autocompletion=service_completer)):
    "Show information about installed service."
    svc: ServiceInfo = None
    try:
        svc = service_registry.get(UUID(service_id))
    except Exception: # pylint: disable=W0703
        svc = service_registry.get_by_name(service_id)
    if svc is None:
        console.print_error('Service not registered!')
        return

    api = []
    for uid in svc.api:
        inf: ServiceInfo = oid_registry.get(uid)
        api.append(str(uid) if inf is None else inf.name)

    vendor = oid_registry.get(svc.vendor)
    vendor = Text(str(svc.vendor), style='yellow') if vendor is None else Text(vendor.description)

    table = Table.grid()
    table.add_column(style='green')
    table.add_column()
    table.add_row('UID:', Text(str(svc.uid), style='yellow'))
    table.add_row('Name:', Text(svc.name))
    table.add_row('Version:', Text(svc.version, style='cyan'))
    table.add_row('Vendor:', vendor)
    table.add_row('Classification: ', Text(svc.classification))
    table.add_row('Description:', Text(svc.description))
    table.add_row('Facilities:', Text(', '.join(svc.facilities)))
    table.add_row('API:', Text(', '.join(api), style='yellow'))
    table.add_row('Distribution:', Text(svc.distribution))
    console.print(table)

@app.command()
def list_applications(with_name: str= \
                      typer.Option('',help="List only applications with this string in name")):
    "Lists installed Saturnin applications."
    apps = list(application_registry.filter(lambda x: with_name in x.name))
    apps.sort(key=attrgetter('name'))
    if apps:
        table = Table(title='Registered applications' if not with_name
                      else f"Registered applications with name containing '{with_name}'",
                      box=box.ROUNDED)
        table.add_column('Application', style='green')
        table.add_column('Version', style='number')
        table.add_column('Used', width=9, justify='center')
        table.add_column('Description')
        for app in apps:
            table.add_row(app.name, app.version, RICH_YES if
                          recipe_registry.any(lambda x: x.application is not None
                                              and x.application == app.uid)
                          else RICH_NO, get_first_line(app.description))
        console.print(table)
    else:
        console.print("There are no Saturnin applications registered.")

@app.command()
def show_application(app_id: str=typer.Argument('', help="Application UID or name",
                                                    autocompletion=application_completer)):
    "Show information about installed application."
    app: ApplicationInfo = None
    try:
        app = application_registry.get(UUID(app_id))
    except Exception: # pylint: disable=W0703
        app = application_registry.get_by_name(app_id)
    if app is None:
        console.print_error('Application not registered!')
        return

    vendor = oid_registry.get(app.vendor)
    vendor = _h(Text(str(app.vendor) if vendor is None else vendor.description))

    table = Table.grid()
    table.add_column(style='green')
    table.add_column()
    table.add_row('UID:', _h(Text(str(app.uid))))
    table.add_row('Name:', Text(app.name))
    table.add_row('Version:', _h(Text(app.version)))
    table.add_row('Vendor:', vendor)
    table.add_row('Classification: ', Text(app.classification))
    table.add_row('Description:', Text(app.description))
    table.add_row('Distribution:', Text(app.distribution))
    console.print(table)

@app.command()
def list_packages():
    """Lists installed distribution packages with Saturnin components.

    This command does not list ALL packages installed in Saturnin virtual environment, but
    only those that contain registered Saturnin components. To list all installed packages,
    use: **pip list**.
    """
    packages = set(service_registry.report('item.distribution'))
    packages.update(application_registry.report('item.distribution'))
    if packages:
        table = Table(title='Installed Saturnin packages', box=box.ROUNDED)
        table.add_column('Package', style='green')
        table.add_column('Version', style='number')
        for dist in (distribution(pkg) for pkg in packages):
            table.add_row(dist.metadata['name'], dist.version)
        console.print(table)
    else:
        console.print("No Saturnin packages are installed.")

@app.command()
def update_registry():
    """Updates registry of installed Saturnin components.

    The registry is updated automatically when Saturnin packages are manipulated with
    built-in **install**, **uninstall** or **pip** commands. Manual update is required only
    when packages are added/updated/removed in differet way.
    """
    console.print('Updating Saturnin service registry ... ', end='')
    try:
        service_registry.clear()
        service_registry.load_from_installed()
        service_registry.save()
        console.print('[ok]OK')
    except Exception: # pylint: disable=W0703
        console.print('[warning]ERROR')
        console.print_exception()
    console.print('Updating Saturnin application registry ... ', end='')
    try:
        application_registry.clear()
        application_registry.load_from_installed()
        application_registry.save()
        console.print('[ok]OK')
    except Exception: # pylint: disable=W0703
        console.print('[warning]ERROR')
        console.print_exception()

@app.command()
def pip(args: List[str]=typer.Argument(None, help="Arguments for pip.")):
    """Runs 'pip' package manager in Saturnin virtual environment.
    """
    pip_cmd = directory_scheme.get_pip_cmd()
    pip_cmd.extend(args)
    if '--help' in args:
        result = subprocess.run(pip_cmd, capture_output=True, text=True) # pylint: disable=W1510
        console.print(_h(Text(pip.__doc__ + result.stdout)))
    else:
        result = subprocess.run(pip_cmd) # pylint: disable=W1510
    if result.returncode != 0:
        console.err_console.print(result.stderr)
    elif '--help' not in args and ('install' in args or 'uninstall' in args):
        update_registry()

@app.command()
def install_package(args: List[str]=typer.Argument(..., help="Arguments for pip install.")):
    """Installs Python package into Saturnin virtual environment via 'pip'.

Note:
   This command is used also to upgrade installed packages using '-U' or '--upgrade' option.
    """
    pip_cmd = directory_scheme.get_pip_cmd('install')
    pip_cmd.extend(args)
    if '--help' in args:
        result = subprocess.run(pip_cmd, capture_output=True, text=True) # pylint: disable=W1510
        text = install_package.__doc__ + sub(r'(\bpip install\b)', 'install package',
                                             result.stdout)
        text = sub(r'(?s)\b(?:Description:).+(?:Install Options:)', "Install Options:",
                   text)
        console.print(_h(Text(text)))
        return None
    #
    with Progress(SpinnerColumn(), TextColumn("{task.description}")) as progress:
        progress.add_task("Installing...", total=1)
        result = subprocess.run(pip_cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                                text=True)
    if result.returncode != 0:
        console.err_console.print(result.stdout)
    update_registry()
    return RESTART

@app.command()
def uninstall_package(args: List[str]=typer.Argument(..., help="Arguments for pip uninstall.")):
    """Uninstalls Python package from Saturnin virtual environment via `pip`.
    """
    pip_cmd = directory_scheme.get_pip_cmd('uninstall')
    pip_cmd.append('--yes')
    pip_cmd.extend(args)
    if '--help' in args:
        result = subprocess.run(pip_cmd, capture_output=True, text=True)
        text = uninstall_package.__doc__ + sub(r'(\bpip uninstall\b)', 'uninstall package',
                                               result.stdout)
        text = sub(r'(?s)\b(?:Description:).+(?:Uninstall Options:)', "Uninstall Options:",
                   text)
        console.print(_h(Text(text)))
        return None
    #
    with Progress(SpinnerColumn(), TextColumn("{task.description}")) as progress:
        progress.add_task("Uninstalling...", total=1)
        result = subprocess.run(pip_cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                                text=True)
    if result.returncode != 0:
        console.err_console.print(result.stdout)
    update_registry()
    return RESTART



# SPDX-FileCopyrightText: 2023-present The Firebird Projects <www.firebirdsql.org>
#
# SPDX-License-Identifier: MIT
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/component/apps.py
# DESCRIPTION:    Application registration and discovery
# CREATED:        23.2.2023
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2023 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""Saturnin application registration and discovery.

Applications are registered as entry points for their `.ApplicationDescriptor` - i.e.
the instance of `.ApplicationDescriptor` for installed service is returned by `.EntryPoint.load()`.

The entry point group for service registration is `saturnin.application`.
"""

from __future__ import annotations
from typing import Dict, Hashable, Optional, Any
from uuid import UUID
from contextlib import suppress
from toml import dumps, loads
from firebird.base.types import Distinct, load
from firebird.base.collections import Registry
from saturnin.base import directory_scheme, ApplicationDescriptor, Error
from saturnin.lib.metadata import get_entry_point_distribution, iter_entry_points

class ApplicationInfo(Distinct): # pylint: disable=R0902
    """Information about application stored in  `.ApplicationRegistry`.

    Arguments:
        uid: Application UID
        name: Application name
        version: Application version
        vendor: Application vendor UID
        classification: Application classification
        description: Application description
        factory: Application factory specification (entry point)
        config: Application configuration factory (entry point)
        descriptor: Application descriptor specification (entry point)
        distribution: Installed distribution package that contains this application
    """
    def __init__(self, *, uid: UUID, name: str, version: str, vendor: UUID,
                 classification: str, description: str, factory: str, config: str,
                 descriptor: str, distribution: str):
        self.__desc_obj: Any = None
        self.__fact_obj: Any = None
        self.__conf_obj: Any = None
        #: Application UID
        self.uid: UUID = uid
        #: Application name
        self.name: str = name
        #: Application version
        self.version: str = version
        #: Application vendor UID
        self.vendor: UUID = vendor
        #: Application classification
        self.classification: str = classification
        #: Application description
        self.description: str = description
        #: Application command factory specification (entry point)
        self.factory: str = factory
        #: Application configuration factory (entry point)
        self.config: str = config
        #: Application descriptor specification (entry point)
        self.descriptor: str = descriptor
        #: Installed distribution package that contains this application
        self.distribution: str = distribution
    def get_key(self) -> Hashable:
        "Returns service UID"
        return self.uid
    def as_toml_dict(self) -> Dict:
        """Returns dictionary with instance data suitable for storage in TOML format
        (values that are not of basic type are converted to string).
        """
        return {'uid': str(self.uid),
                'name': self.name,
                'version': self.version,
                'vendor': str(self.vendor),
                'classification': self.classification,
                'description': self.description,
                'factory': self.factory,
                'config': self.config,
                'descriptor': self.descriptor,
                'distribution': self.distribution,
                }
    def get_recipe_name(self) -> str:
        """Returns default recipe name for this application. If application name contains
        dots, only part after last dot is returned. Otherwise it returns the application name.
        """
        return self.name.split('.')[-1]
    @property
    def descriptor_obj(self) -> ApplicationDescriptor:
        """Application descriptor object. If it's not assigned directly, then it's loaded
        using `.descriptor` on first access.
        """
        if self.__desc_obj is None:
            self.__desc_obj = load(self.descriptor)
        return self.__desc_obj
    @descriptor_obj.setter
    def set_descriptor_obj(self, value: Optional[ApplicationDescriptor]) -> None:
        "Property setter"
        self.__desc_obj = value
    @property
    def factory_obj(self) -> Any:
        """Application command factory object. If it's not assigned directly, then it's loaded
        using `.factory` on first access.
        """
        if self.__fact_obj is None:
            self.__fact_obj = load(self.factory)
        return self.__fact_obj
    @factory_obj.setter
    def set_factory_obj(self, value: Optional[Any]) -> None:
        "Property setter"
        self.__fact_obj = value
    @property
    def config_obj(self) -> Any:
        """Application configuration factory object. If it's not assigned directly, then it's loaded
        using `.config` on first access.
        """
        if self.__conf_obj is None:
            self.__conf_obj = load(self.config)
        return self.__conf_obj
    @factory_obj.setter
    def set_config_obj(self, value: Optional[Any]) -> None:
        "Property setter"
        self.__conf_obj = value

class ApplicationRegistry(Registry):
    """Saturnin application registry.

    Holds `.ApplicationInfo` instances.

    It is used in two modes:

    1. In full saturnin deployment, the information about applications is loaded from TOML file.
       Application descriptors and factories are loaded on demand.
    2. In standalone service/bundle mode, application information including app. desciptors
       and factories is stored directly by executor script, so there is no dynamic discovery
       and the whole could be compiled with Nutika.
    """
    def add(self, descriptor: ApplicationDescriptor, factory: Any, distribution: str) -> None:
        """Direct application registration. Used by systems that does not allow dynamic discovery,
        for example programs compiled by Nuitka.

        Arguments:
            descriptor: Application descriptor
            factory: Application factory
            distribution: Distribution package name with application
        """
        kwargs = {}
        kwargs['distribution'] = distribution
        kwargs['descriptor'] = ''
        kwargs['uid'] = descriptor.uid
        kwargs['name'] = descriptor.name
        kwargs['version'] = descriptor.version
        kwargs['vendor'] = descriptor.vendor_uid
        kwargs['classification'] = descriptor.classification
        kwargs['description'] = descriptor.description
        kwargs['factory'] = descriptor.factory
        kwargs['config'] = descriptor.config
        app_info = ApplicationInfo(**kwargs) # pylint: disable=E1125
        app_info.descriptor_obj = descriptor
        app_info.factory_obj = factory
        self.store(app_info)
    def load_from_installed(self, *, ignore_errors: bool=False) -> None:
        """Populate registry from descriptors of installed applications.

        Arguments:
          ignore_errors: When True, errors are ignored, otherwise `.Error` is raised.
        """
        for entry in iter_entry_points('saturnin.application'):
            kwargs = {}
            dist = get_entry_point_distribution(entry)
            kwargs['distribution'] = dist if dist is None else dist.metadata['name']
            kwargs['descriptor'] = entry.value
            try:
                desc: ApplicationDescriptor = entry.load()
            except Exception as exc:
                if ignore_errors:
                    continue
                raise Error(f"Failed to load application '{entry.name}' "
                            f"from '{kwargs['distribution']}'") from exc
            kwargs['uid'] = desc.uid
            kwargs['name'] = desc.name
            kwargs['version'] = desc.version
            kwargs['vendor'] = desc.vendor_uid
            kwargs['classification'] = desc.classification
            kwargs['description'] = desc.description
            kwargs['factory'] = desc.factory
            kwargs['config'] = desc.config
            try:
                app_info = ApplicationInfo(**kwargs) # pylint: disable=E1125
            except Exception as exc:
                if ignore_errors:
                    continue
                raise Error(f"Malformed application descriptor for '{entry.name}' "
                            f"from '{kwargs['distribution']}'") from exc
            with suppress(ValueError):
                self.store(app_info)
    def load_from_toml(self, toml: str, *, ignore_errors: bool=False) -> None:
        """Populate registry from TOML document.

        Arguments:
          toml: TOML document (as created by `as_toml` method).
          ignore_errors: When True, errors are ignored, otherwise `.Error` is raised.
        """
        data = loads(toml)
        self.clear()
        for uid, kwargs in data.items():
            try:
                kwargs['uid'] = UUID(kwargs['uid'])
                kwargs['vendor'] = UUID(kwargs['vendor'])
                app_info = ApplicationInfo(**kwargs)
            except Exception as exc:
                if ignore_errors:
                    continue
                raise Error(f"Malformed application data for '{uid}' "
                            f"from '{kwargs['distribution']}'") from exc
            self.store(app_info)
    def as_toml(self) -> str:
        """Returns registry content as TOML document.
        """
        nodes = {str(node.uid): node.as_toml_dict() for node in self._reg.values()}
        return dumps(nodes)
    def load(self) -> None:
        "Read information about installed applications from previously saved TOML file."
        if directory_scheme.site_apps_toml.is_file():
            application_registry.load_from_toml(directory_scheme.site_apps_toml.read_text())
    def save(self) -> None:
        "Save information about installed applications to TOML file."
        directory_scheme.site_apps_toml.write_text(application_registry.as_toml())
    def get_by_name(self, name: str, default: Any=None) -> Distinct:
        """Get application by its name.

        Arguments:
            name: Application name.
            default: Default value returned when application is not found.
        """
        return self.find(lambda x: x.name == name, default=default)

#: Saturnin application registry
application_registry: ApplicationRegistry = ApplicationRegistry()
application_registry.load()



# SPDX-FileCopyrightText: 2020-present The Firebird Projects <www.firebirdsql.org>
#
# SPDX-License-Identifier: MIT
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/_scripts/bundlerun.py
# DESCRIPTION:    Script to run bundle of services
# CREATED:        6.12.2020
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2020 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""Saturnin script to run bundle of services.


"""

from __future__ import annotations
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter, Action
from configparser import ConfigParser, ExtendedInterpolation
import logging
from logging.config import fileConfig
from firebird.base.logging import get_logger, Logger, ANY, bind_logger
from firebird.base.trace import trace_manager
from saturnin.base import directory_scheme, SECTION_BUNDLE
from saturnin.component.controller import Outcome
from saturnin.component.bundle import BundleExecutor

LOG_FORMAT = '%(levelname)s [%(processName)s/%(threadName)s] %(message)s'

class UpperAction(Action):
    """Converts argument to uppercase.
    """
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values.upper())

def main(description: str=None, bundle_config: str=None):
    """Saturnin script to run bundle of services.

    Arguments:
      description: Description shown when `--help` is used.
      bundle_config: Default value for `BUNDLE-CONFIG` argument.

    usage::

      saturnin-bundle [-h] [-c CONFIG] [-s SECTION] [-q] [-o]
                      [-l {critical,fatal,error,warn,warning,info,debug,notset}]
                      BUNDLE-CONFIG

    positional arguments:
      BUNDLE-CONFIG         Path to service bundle configuration file.

    optional arguments:
      -h, --help            show this help message and exit
      -c CONFIG, --config CONFIG
                            Path to additional configuration file. Could be specified multiple times. (default: None)
      -s SECTION, --section SECTION
                            Configuration section name (default: bundle)
      -q, --quiet           Suppress console output. (default: False)
      -o, --outcome         Always print service execution outcome (default: False)
      -l {critical,fatal,error,warn,warning,info,debug,notset}, --log-level {critical,fatal,error,warn,warning,info,debug,notset}
                            Logging level (default: None)
    """
    if description is None:
        description = "Saturnin script to run bundle of services."
    parser: ArgumentParser = ArgumentParser(description=description,
                                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument('service', metavar='BUNDLE-CONFIG',
                        help="Path to service bundle configuration file.",
                        nargs=1 if bundle_config is None else '?',
                        default=bundle_config)
    parser.add_argument('-c','--config', metavar='CONFIG', action='append',
                        help="Path to additional configuration file. "
                             "Could be specified multiple times.")
    parser.add_argument('-s', '--section', help="Configuration section name",
                        default=SECTION_BUNDLE)
    parser.add_argument('-q', '--quiet', action='store_true', help="Suppress console output.",
                        default=False)
    parser.add_argument('-o','--outcome', action='store_true',
                        help="Always print service execution outcome", default=False)
    parser.add_argument('-l', '--log-level', action=UpperAction,
                        choices=[x.lower() for x in logging._nameToLevel
                                 if isinstance(x, str)],
                        help="Logging level")

    args = parser.parse_args()

    main_config: ConfigParser = ConfigParser(interpolation=ExtendedInterpolation())
    cfg_files = [str(directory_scheme.logging_conf)]
    if args.config:
        cfg_files.extend(args.config)
    if isinstance(args.service, list):
        cfg_files.extend(args.service)
    else:
        cfg_files.append(args.service)
    cfg_files = main_config.read(cfg_files)
    # Logging configuration
    if main_config.has_section('loggers'):
        fileConfig(main_config)
    else:
        logging.basicConfig(format=LOG_FORMAT)
    bind_logger(ANY, ANY, 'saturnin')
    log: Logger = get_logger('saturnin')
    if args.log_level is not None:
        log.setLevel(args.log_level)
    # trace configuration
    if main_config.has_section('trace'):
        trace_manager.load_config(main_config)

    try:
        with BundleExecutor('saturnin-bundle') as executor:
            executor.configure(cfg_files, section=args.section)
            result = executor.run()
            if not args.quiet:
                for name, outcome, details in result:
                    if outcome is not Outcome.OK or args.outcome:
                        print(f'{name}: {outcome.value}')
                        if details:
                            for line in details:
                                print(f' {line}')
    except Exception as exc: # pylint: disable=W0703
        log.exception("Service execution failed")
        parser.exit(1, f'{exc!s}\n')
    finally:
        logging.shutdown()

if __name__ == '__main__':
    main()
